#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

""" 
This script will monitors the pre-processing progress in CryoSparc and notify to EMhub.
"""

import sys, os
import time
import datetime as dt
from glob import glob
import argparse
from pprint import pprint
from contextlib import contextmanager
import numpy as np

import mrcfile


from emhub.client import open_client
from emhub.utils.image import Base64Converter


def usage(error):
    print("ERROR: %s" % error)
    get_parser().print_usage()
    sys.exit(1)


def get_parser():
    """ Return the argparse parser, so we can get the arguments """

    parser = argparse.ArgumentParser()
    add = parser.add_argument  # shortcut

    g = parser.add_mutually_exclusive_group()
    g.add_argument('--list', action='store_true',
                   help="List existing sessions in the server.")
    g.add_argument('--update', action='store_true',
                   help="Update an existing session")

    add('-p', '--project', metavar='PROJECT_PATH',
        help='Project path.')

    add('--session_id', type=int,
        help='EMhub Session Id')

    # add('--prot_ctf', type=int, default=0,
    #     help="Id of the CTF protocol to monitor. ")
    #
    # add('--prot_picking', type=int, default=0,
    #     help="Id of the particle picking protocol to monitor. ")
    #
    # add('--prot_2d', type=int, default=0,
    #     help='Id of the classify 2d protocol.')

    add('--clear', action='store_true',
        help="Clear existing data associated with the session.")

    return parser


class CSLiveSession:
    def __init__(self, projPath, sessionId):
        self._sessionId = sessionId

        if not os.path.exists(projPath):
            usage("No project found at '%s'" % projPath)

        self._projectPath = projPath

    def _get_path(self, *paths):
        return os.path.join(self._projectPath, *paths)

    def _update_mics_ctfs(self, dc):
        files = glob(self._get_path('import_movies', '*.tiff'))
        self._micSetId = 'Micrographs_%06d' % 1  # FIXME: give a proper id

        attrs = {
            'session_id': self._sessionId,
            'set_id': self._micSetId
        }

        stats = {
            'numOfCls2D': 0,
            'numOfCtfs': 0,
            'numOfMics': 0,
            'numOfMovies': 0,
            'numOfPtcls': 0,
            'ptclSizeMax': 0,
            'ptclSizeMin': 0
        }

        micBase64 = Base64Converter()
        psdBase64 = Base64Converter()

        idset = set()
        idcount = 0
        mics_count = 0
        parts_count = 0

        def _get(folder, movieRoot, suffix):
            pattern = self._get_path(folder, movieRoot + '*' + suffix)
            matched = glob(pattern)
            print('pattern: ', pattern, 'matched: ', len(matched))
            return matched[0] if matched else None

        for fn in files:
            movieRoot, movieExt = os.path.splitext(os.path.basename(fn))
            extractFn = _get('extract', movieRoot, '.cs')
            micArray = np.load(extractFn)
            row = micArray[0]
            micId = int(row['location/micrograph_uid'])

            try:
                u = float(row['ctf/df1_A'])
                v = float(row['ctf/df2_A'])
                a = float(row['ctf/df_angle_rad'])
            except:
                print('Missing CTF value for Micrograph id "%s", ignoring.' % micId)
                continue

            mic = row['location/micrograph_path'].decode("utf-8")
            pixelSize = float(row['blob/psize_A'])

            if micId in idset:
                print('Micrograph id "%s" seems duplicated, ignoring.' % micId)
                continue

            idcount += 1
            attrs.update({
                'item_id': idcount,
                'uid': str(micId),
                'ctfDefocus': (u + v) * 0.5,
                'ctfDefocusU': u,
                'ctfDefocusV': v,
                'ctfDefocusAngle': np.rad2deg(a),
                'ctfResolution': 0.0,
                'ctfFit': 0.0,
                'location': mic,
                'ctfFitData': '',
                'shiftPlotData': '',
                'pixelSize': pixelSize
            })

            print("\nAdding item %06d: \n   -> %s" % (micId, mic))
            psdPath = _get('ctfestimated', movieRoot, 'diag_2D.mrc')

            print("  PSD: ", psdPath, "exists: ", os.path.exists(psdPath))
            if os.path.exists(psdPath):
                attrs['psdData'] = psdBase64.from_mrc(psdPath)

            micPath = self._get_path(mic.replace('S1/', ''))
            print("  MIC: ", micPath, "exists: ", os.path.exists(micPath))
            if os.path.exists(micPath):
                attrs['micThumbData'] = micBase64.from_mrc(micPath)
                attrs['micThumbPixelSize'] = pixelSize * micBase64.scale
                w, h = row['location/micrograph_shape']
                attrs['coordinates'] = [
                    (r['location/center_x_frac'] * h,
                     r['location/center_y_frac'] * w) for r in micArray]

            try:
                dc.add_session_item(attrs)
                idset.add(micId)
                mics_count += 1
                parts_count += len(attrs.get('coordinates', []))
            except Exception as e:
                print(">>> FAILED: id exists: ", micId in idset)
                pprint(attrs)
                raise e

            stats = {
                'numOfCls2D': 0,
                'numOfCtfs': mics_count,
                'numOfMics': mics_count,
                'numOfMovies': mics_count,
                'numOfPtcls': parts_count,
                'ptclSizeMax': 0,
                'ptclSizeMin': 0
            }

            if mics_count % 10 == 0:
                dc.update_session({'id': self._sessionId, 'stats': stats})
                time.sleep(10)  # FIXME

        dc.update_session({'id': self._sessionId, 'stats': stats})

    def _update_coords(self, dc):
        protCtf = self._protocols['ctf']
        outputCTF = getattr(protCtf, 'outputCTF', None)
        micSet = protCtf.inputMicrographs.get()

        protPicking = self._protocols.get('picking', None)

        if protPicking is None:
            return

        coordsSet = getattr(protPicking, 'outputCoordinates', None)
        attrs = {
            'session_id': self._sessionId,
            'set_id': 'Micrographs_%06d' % micSet.getObjId()
        }
        coordList = []
        lastMicId = None

        for coord in coordsSet.iterItems(orderBy='_micId', direction='ASC'):
            micId = coord.getMicId()
            if micId != lastMicId:
                # Update the coordinates information if necessary
                if lastMicId is not None:
                    attrs['item_id'] = lastMicId
                    attrs['coordinates'] = coordList

                    dc.update_session_item(attrs)
                lastMicId = micId
                coordList = []
            coordList.append(coord.getPosition())

    def run(self):
        with open_client() as dc:
            sessionDict = dc.get_session(self._sessionId)
            self._update_mics_ctfs(dc)


def main():
    args = get_parser().parse_args()

    if args.clear:
        with open_client() as dc:
            print("Clearing session data: ")
            dc.request('clear_session_data', jsonData={'attrs': {'id': args.session_id}})

    if args.list:
        with open_client() as dc:
            r = dc.request('get_sessions', jsonData={})

            for session in r.json():
                pprint(session)

    elif args.project:
        CSLiveSession(args.project, args.session_id).run()

    else:
        print("Please provide some arguments")
        sys.exit(1)


if __name__ == '__main__':
    main()








# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *              Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [2]
# *
# * [1] SciLifeLab, Stockholm University
# * [2] MRC Laboratory of Molecular Biology (MRC-LMB)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

import datetime as dt


class TestDataBase:
    """ Base class to create a dataset for a given DataManager.
    """

    TZ_DELTA = 0  # Define timezone, UTC '0'
    tzinfo = dt.timezone(dt.timedelta(hours=TZ_DELTA))

    def __init__(self, dm):
        """
        Args:
            dm: DataManager with db to create test data
        """
        dm.create_admin()
        self._populateTestData(dm)

    def _populateTestData(self, dm):
        # Create tables with test data for each database model
        print("Populating forms...")
        self._populateForms(dm)
        print("Populating users...")
        self._populateUsers(dm)
        print("Populating resources...")
        self._populateResources(dm)
        print("Populating applications...")
        self._populateApplications(dm)
        print("Populating bookings...")
        self._populateBookings(dm)
        print("Populating sessions...")
        self._populateSessions(dm)

    def _populateForms(self, dm):
        form1 = {
            'title': 'Sample Form',
            'params': [
                {'id': 'name',
                 'label': 'Name',
                 },
                {'id': 'phone',
                 'default': '88546756777',
                 'label': 'Phone',
                 },
                {'id': 'show-phone',
                 'label': 'Show phone?',
                 'type': 'bool'
                 },
                {'id': 'about',
                 'label': 'About everything that you need',
                 'type': 'text'
                 },
                {'label': 'Options'},
                {'id': 'level',
                 'label': 'Level',
                 'enum': {'choices': ['low', 'medium', 'high'],
                          'display': 'combo'
                          }
                 },
                {'id': 'pet',
                 'label': 'Pet',
                 'enum': {'choices': ['cat', 'dog', 'horse', 'monkey'],
                          'display': 'radio'
                          }
                 },
        ]}

        dm.create_form(name='sample',
                       definition=form1)

        form2 = {
            'title': 'Experiment',
            'sections': [
                {'label': 'Basic',
                 'params': [
                     {'id': 'grid_prep_needed',
                      'label': 'Grids preparation needed?',
                      'type': 'bool'
                      },
                     {'id': 'grid_ready_screen',
                      'label': 'Grids ready to be screened?',
                      'type': 'bool'
                      },
                     {'id': 'data_collection',
                      'label': 'Data collection?',
                      'type': 'bool'
                      },
                     {'id': 'grid_clipped',
                      'label': 'Are grids clipped?',
                      'type': 'bool'
                      },
                     {'id': 'grid_type',
                      'label': 'Grid type',
                      },
                 ]},
                {'label': 'Grid Location',
                 'params': [
                     {'id': 'dewar_number',
                      'label': 'Dewar number',
                      },
                     {'id': 'cane_number',
                      'label': 'Cane number',
                      },
                     {'id': 'puck_number',
                      'label': 'Puck number (color/name)',
                      },
                     {'id': 'gridbox_number',
                      'label': 'Grid box number',
                      },
                     {'id': 'gridbox_label',
                      'label': 'Grid box label',
                      },
                     {'id': 'slot_numbers',
                      'label': ' Slot numbers',
                      },
                 ]},
                {'label': 'Detector',
                 'params': [
                     {'id': 'detector_mode',
                      'label': 'Detector mode',
                      'enum': {'choices': ['linear', 'counting'],
                               'display': 'radio'
                               }
                      },
                     {'id': 'pixel_size',
                      'label': 'Pixel Size (Å)',
                      },
                     {'id': 'dose_rate',
                      'label': 'Dose Rate (e/px/s)',
                      },
                     {'id': 'total_dose',
                      'label': 'Total dose (e/Å2)',
                      },
                     {'id': 'defocuses',
                      'label': 'Defocuses (um)',
                      },
                 ]},
                {'label': 'Other',
                 'params': [
                     {'id': 'screening_instructions',
                      'help': '(for instance protein concentration, blotting '
                              'time and other useful info)',
                      'label': 'Screening instructions',
                      'type': 'text'
                      },
                     {'id': 'sample_information',
                      'help': '(Mw, Dimensions (Å), multimerisation state, etc.)',
                      'label': 'Sample Information',
                      'type': 'text'
                      },
                     {'id': 'aim_experiment',
                      'label': 'Aim of experiment',
                      'type': 'text'
                      },
                 ]},
            ]
        }

        dm.create_form(name='experiment',
                       definition=form2)

    def _populateUsers(self, dm):
        pass

    def _populateResources(self, dm):
        resources = [
            {'name': 'Solna Krios α', 'tags': 'microscope krios solna',
             'image': 'titan-krios.png', 'color': 'rgba(58, 186, 232, 1.0)',
             'extra': {'latest_cancellation': 48,
                       'requires_slot': True,
                       'min_booking': 8,
                       'max_booking': 72}},
            {'name': 'Solna Krios β', 'tags': 'microscope krios solna',
             'status': 'inactive',
             'image': 'titan-krios.png', 'color': 'rgba(60, 90, 190, 1.0)',
             'extra': {'latest_cancellation': 48,
                       'requires_slot': True,
                       'min_booking': 8,
                       'max_booking': 72}},
            {'name': 'Talos', 'tags': 'microscope talos solna',
             'image': 'talos-artica.png', 'color': 'rgba(43, 84, 36, 1.0)',
             'extra': {'latest_cancellation': 48,
                       'requires_slot': True,
                       'min_booking': 8,
                       'max_booking': 72}},
            {'name': 'Vitrobot 1', 'tags': 'instrument solna',
             'image': 'vitrobot.png', 'color': 'rgba(158, 142, 62, 1.0)'},
            {'name': 'Vitrobot 2', 'tags': 'instrument solna',
             'image': 'vitrobot.png', 'color': 'rgba(69, 62, 25, 1.0)'},
            {'name': 'Carbon Coater', 'tags': 'instrument solna',
             'image': 'carbon-coater.png', 'color': 'rgba(48, 41, 40, 1.0)'},
            {'name': 'Users Drop-in', 'tags': 'service solna',
             'image': 'users-dropin.png', 'color': 'rgba(68, 16, 105, 1.0)',
             'extra': {'requires_slot': True}},

            # Umeå instruments
            {'name': 'Umeå Krios', 'tags': 'microscope krios umea',
             'image': 'titan-krios.png', 'color': 'rgba(15, 40, 130, 1.0)',
             'extra': {'latest_cancellation': 48,
                       'requires_slot': True,
                       'min_booking': 8,
                       'max_booking': 72}},
        ]

        for rDict in resources:
            dm.create_resource(**rDict)

    def _populateApplications(self, dm):
        pass

    def _populateBookings(self, dm):
        pass

    def _populateSessions(self, dm):
        pass

    def _datetime(self, *args):
        return dt.datetime(*args, tzinfo=self.tzinfo)



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

import threading
import flask_mail


class MailManager:
    """ Helper class to send emails via Flask app. """
    def __init__(self, app):
        self._app = app
        self._mail = flask_mail.Mail(app)

    def __send_async(self, msg):
        with self._app.app_context():
            self._mail.send(msg)

    def send_mail(self, recipients, subject, body, sender=None, html_body=None):
        sender = sender or self._app.config['MAIL_DEFAULT_SENDER']
        msg = flask_mail.Message(subject, sender=sender, recipients=recipients)
        msg.body = body
        msg.html = html_body
        threading.Thread(target=self.__send_async, args=(msg,)).start()



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *              Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [2]
# *
# * [1] SciLifeLab, Stockholm University
# * [2] MRC Laboratory of Molecular Biology (MRC-LMB)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

import json
import unittest
import random
import datetime as dt

from emhub.client import DataClient
from emhub.utils import (get_quarter, pretty_quarter)


class TestClientApi(unittest.TestCase):
    def test_create_user(self):
        """ Create a user and check the result. """
        username = "queen%d" % random.randint(1, 1000)
        sc = DataClient()
        sc.login('mull', 'mull')
        method = "create_user"
        jsonData = {"attrs":{"username":"%s" % username,
                             "email":"%s@emhub.org" % username,
                             "name":"Queen of England",
                             "password":"grgrh45$%^4573",
                             "roles":"user",
                             "phone":"343-332-4525"}}

        print("="*80, "\nCreating user: %s" % jsonData)
        sc.request(method, jsonData=jsonData)
        result_id = json.loads(sc.json())['user']['id']
        print("Created new user with id: %s" % result_id)

        sc.request(method="get_users",
                   jsonData={"condition":"id=%s" % result_id})
        check = json.loads(sc.json())[0]['id']

        self.assertEqual(result_id, check, msg="Creating test user failed!")

        sc.logout()

    def test_create_session(self):
        """ Create a session and check the result. """
        name = "mysession_%d" % random.randint(1, 1000)
        sc = DataClient()
        sc.login('mull', 'mull')
        method = "create_session"
        jsonData = {"attrs": {"name":"%s" % name,
                              "resource_id":"2",
                              "operator_id":"23"}}

        print("="*80, "\nCreating session: %s" % jsonData)
        sc.request(method, jsonData=jsonData)
        result_id = json.loads(sc.json())['session']['id']
        print("Created new session with id: %s" % result_id)

        sc.request(method="get_sessions",
                   jsonData={"condition":"id=%s" % result_id})
        check = json.loads(sc.json())[0]['id']

        self.assertEqual(result_id, check, msg="Creating test session failed!")

        sc.logout()

    def test_create_invoice_periods(self):
        q1 = get_quarter()
        q0 = get_quarter(q1[0] - dt.timedelta(days=1))

        sc = DataClient()
        sc.login('mull', 'mull')
        method = "create_invoice_period"

        for q in [q0, q1]:
            pass

        sc.logout()


# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *              Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [2]
# *
# * [1] SciLifeLab, Stockholm University
# * [2] MRC Laboratory of Molecular Biology (MRC-LMB)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

import os
import time
import json
from glob import glob

import flask
from flask import request
from flask import current_app as app
import flask_login

from emhub.utils import (datetime_from_isoformat, datetime_to_isoformat,
                         send_json_data, send_error)
from emhub.data import DataContent


api_bp = flask.Blueprint('api', __name__)


# ---------------------------- AUTH  ------------------------------------------

@api_bp.route('/login', methods=['POST'])
def login():
    username = request.json['username']
    password = request.json['password']

    user = app.dm.get_user_by(username=username)
    if user is None or not user.check_password(password):
        return send_error('Invalid username or password')

    flask_login.login_user(user)

    return send_json_data('OK')


@api_bp.route('/logout', methods=['POST'])
@flask_login.login_required
def logout():
    flask_login.logout_user()

    return send_json_data('OK')


# ---------------------------- USERS ------------------------------------------

@api_bp.route('/create_user', methods=['POST'])
@flask_login.login_required
def create_user():
    return handle_user(app.dm.create_user)


@api_bp.route('/register_user', methods=['POST'])
@flask_login.login_required
def register_user():
    def register(**attrs):
        user = app.dm.create_user(
            username=attrs['email'],
            email=attrs['email'],
            phone='',
            password=os.urandom(24).hex(),
            name=attrs['name'],
            roles=attrs['roles'],
            pi_id=attrs['pi_id'],
            status='pending'
        )
        app.mm.send_mail(
            [user.email],
            "emhub: New account registered",
            flask.render_template('email/account_registered.txt',
                                  user=user))
        return user

    return handle_user(register)


@api_bp.route('/update_user', methods=['POST'])
@flask_login.login_required
def update_user():
    return handle_user(app.dm.update_user)


@api_bp.route('/delete_user', methods=['POST'])
@flask_login.login_required
def delete_user():
    return handle_user(app.dm.delete_user)


@api_bp.route('/update_user_form', methods=['POST'])
@flask_login.login_required
def update_user_form():
    try:
        f = request.form

        attrs = {'id': f['user-id'],
                 'name': f['user-name'],
                 'phone': f['user-phone'],
                 'status': f['user-status-select']
                 }

        roles = [v.replace('role-', '') for v in f if v.startswith('role-')]
        if roles:
            attrs['roles'] = roles

        if 'user-pi-select' in f:
            pi_id = int(f['user-pi-select'])
            if pi_id:
                attrs['pi_id'] = pi_id
            # TODO: Validate if a user is not longer PI
            # check that there are not other users referencing this one as pi
            # still this will not be a very common case

        password = f['user-password'].strip()
        if password:
            attrs['password'] = password

        if 'user-profile-image' in request.files:
            profile_image = request.files['user-profile-image']

            if profile_image.filename:
                _, ext = os.path.splitext(profile_image.filename)

                if ext.lstrip(".").upper() not in app.config["ALLOWED_IMAGE_EXTENSIONS"]:
                    return send_error("Image format %s is not allowed!" % ext.upper())
                else:
                    image_name = 'profile-image-%06d%s' % (int(f['user-id']), ext)
                    image_path = os.path.join(app.config['USER_IMAGES'], image_name)
                    profile_image.save(image_path)
                    attrs['profile_image'] = image_name

        app.dm.update_user(**attrs)

        return send_json_data({'user': attrs})

    except Exception as e:
        print(e)
        return send_error('ERROR from Server: %s' % e)


@api_bp.route('/get_users', methods=['POST'])
@flask_login.login_required
def get_users():
    return filter_request(app.dm.get_users)


# ---------------------------- APPLICATIONS -----------------------------------

@api_bp.route('/create_template', methods=['POST'])
@flask_login.login_required
def create_template():
    return handle_template(app.dm.create_template)


@api_bp.route('/get_templates', methods=['POST'])
@flask_login.login_required
def get_templates():
    return filter_request(app.dm.get_templates)


@api_bp.route('/update_template', methods=['POST'])
@flask_login.login_required
def update_template():
    return handle_template(app.dm.update_template)


@api_bp.route('/delete_template', methods=['POST'])
@flask_login.login_required
def delete_template():
    return handle_template(app.dm.delete_template)


@api_bp.route('/create_application', methods=['POST'])
@flask_login.login_required
def create_application():
    return handle_application(app.dm.create_application)


@api_bp.route('/get_applications', methods=['POST'])
@flask_login.login_required
def get_applications():
    return filter_request(app.dm.get_applications)


@api_bp.route('/update_application', methods=['POST'])
@flask_login.login_required
def update_application():
    return handle_application(app.dm.update_application)


@api_bp.route('/delete_application', methods=['POST'])
@flask_login.login_required
def delete_application():
    return handle_application(app.dm.delete_application)


@api_bp.route('/import_application', methods=['POST'])
@flask_login.login_required
def import_application():
    try:
        if not request.json:
            raise Exception("Expecting JSON request.")

        orderCode = request.json['code'].upper()

        dm = app.dm

        application = dm.get_application_by(code=orderCode)

        if application is not None:
            raise Exception('Application %s already exist' % orderCode)

        orderJson = app.sll_pm.fetchOrderDetailsJson(orderCode)

        if orderJson is None:
            raise Exception('Invalid application ID %s' % orderCode)

        piEmail = orderJson['owner']['email'].lower()
        # orderId = orderJson['identifier']

        pi = dm.get_user_by(email=piEmail)

        if pi is None:
            raise Exception("Order owner email (%s) not found as PI" % piEmail)

        if orderJson['status'] not in ['accepted', 'processing']:
            raise Exception("Only applications with status 'accepted' or "
                            "'processing' can be imported. ")

        fields = orderJson['fields']
        description = fields.get('project_des', None)
        invoiceRef = fields.get('project_invoice_addess', None)

        created = datetime_from_isoformat(orderJson['created'])
        pi_list = fields.get('pi_list', [])

        form = orderJson['form']
        iuid = form['iuid']

        # Check if the given form (here templates) already exist
        # or we need to create a new one
        orderTemplate = None
        templates = dm.get_templates()
        for t in templates:
            if t.extra.get('portal_iuid', None) == iuid:
                orderTemplate = t
                break

        if orderTemplate is None:
            orderTemplate = dm.create_template(
                title=form['title'],
                status='active',
                extra={'portal_iuid': iuid}
            )
            dm.commit()

        application = dm.create_application(
            code=orderCode,
            title=orderJson['title'],
            created=created,  # datetime_from_isoformat(o['created']),
            status='active',
            description=description,
            creator_id=pi.id,
            template_id=orderTemplate.id,
            invoice_reference=invoiceRef or 'MISSING_INVOICE_REF',
        )

        for piTuple in pi_list:
            piEmail = piTuple[1].lower()
            pi = dm.get_user_by(email=piEmail)
            if pi is not None:
                application.users.append(pi)

        dm.commit()

        return send_json_data({'application': application.json()})

    except Exception as e:
        print(e)
        import traceback
        traceback.print_exc()

        return send_error('ERROR from Server: %s' % e)


# ---------------------------- RESOURCES ---------------------------------------

@api_bp.route('/get_resources', methods=['POST'])
@flask_login.login_required
def get_resources():
    return filter_request(app.dm.get_resources)


@api_bp.route('/create_resource', methods=['POST'])
@flask_login.login_required
def create_resource():
    return handle_resource(app.dm.create_resource)

@api_bp.route('/update_resource', methods=['POST'])
@flask_login.login_required
def update_resource():
    return handle_resource(app.dm.update_resource)

@api_bp.route('/delete_resource', methods=['POST'])
@flask_login.login_required
def delete_resource():
    return handle_resource(app.dm.delete_resource)

# ---------------------------- BOOKINGS ---------------------------------------

@api_bp.route('/create_booking', methods=['POST'])
@flask_login.login_required
def create_booking():
    def create(**attrs):
        check_min = request.json.get('check_min_booking', True)
        check_max = request.json.get('check_max_booking', True)
        return app.dm.create_booking(check_min_booking=check_min,
                                     check_max_booking=check_max,
                                     **attrs)

    return handle_booking('bookings_created', create)


@api_bp.route('/get_bookings', methods=['POST'])
@flask_login.login_required
def get_bookings():
    return filter_request(app.dm.get_bookings)


# Shortcut method to get a range of bookings, used by the Calendar
@api_bp.route('/get_bookings_range', methods=['POST'])
@flask_login.login_required
def get_bookings_range():
    d = request.json or request.form
    bookings = app.dm.get_bookings_range(
        datetime_from_isoformat(d['start']),
        datetime_from_isoformat(d['end'])
    )
    func = app.dc.booking_to_event
    return send_json_data([func(b) for b in bookings])


@api_bp.route('/update_booking', methods=['POST'])
@flask_login.login_required
def update_booking():
    return handle_booking('bookings_updated', app.dm.update_booking)


@api_bp.route('/delete_booking', methods=['POST'])
@flask_login.login_required
def delete_booking():
    # When deleting we don't need to send all info back, just ID
    def _transform(b):
        return {'id': b.id}

    return handle_booking('bookings_deleted', app.dm.delete_booking,
                          booking_transform=_transform)


# ---------------------------- SESSIONS ---------------------------------------

@api_bp.route('/get_sessions', methods=['POST'])
@flask_login.login_required
def get_sessions():
    return filter_request(app.dm.get_sessions)


@api_bp.route('/poll_sessions', methods=['POST'])
@flask_login.login_required
def poll_sessions():
    session_folders = app.dm.get_session_folders()

    def _user(u):
        if not u:
            return {}
        else:
            return {'name': u.name,
                    'email': u.email
                    }

    while True:
        sessions = app.dm.get_sessions(condition='status=="pending"')
        if sessions:
            for s in sessions:
                b = s.booking
                e = app.dc.booking_to_event(b)
                data = [{
                    'id': s.id,
                    'name': s.name,
                    'booking_id': s.booking_id,
                    'start': datetime_to_isoformat(s.start),
                    'user': _user(b.owner),
                    'pi': _user(b.owner.get_pi()),
                    'operator': _user(b.operator),
                    'folder': session_folders[s.name[:3]],
                    'title': e['title']
                 }]
                return send_json_data(data)
        time.sleep(3)


@api_bp.route('/create_session', methods=['POST'])
@flask_login.login_required
def create_session():
    return handle_session(app.dm.create_session)


@api_bp.route('/update_session', methods=['POST'])
@flask_login.login_required
def update_session():
    return handle_session(app.dm.update_session)


@api_bp.route('/delete_session', methods=['POST'])
@flask_login.login_required
def delete_session():
    return handle_session(app.dm.delete_session)


@api_bp.route('/load_session', methods=['POST'])
@flask_login.login_required
def load_session():
    return handle_session(app.dm.load_session)


@api_bp.route('/clear_session_data', methods=['POST'])
@flask_login.login_required
def clear_session_data():
    return handle_session(app.dm.clear_session_data)


@api_bp.route('/create_session_set', methods=['POST'])
@flask_login.login_required
def create_session_set():
    """ Create a set file without actual session. """
    def handle(session, set_id, **attrs):
        session.data.create_set(set_id, attrs)
        return {'session_set': {}}

    return handle_session_data(handle, mode="a")


@api_bp.route('/update_session_set', methods=['POST'])
@flask_login.login_required
def update_session_set():
    """ Create a set file without actual session. """
    def handle(session, set_id, **attrs):
        session.data.update_set(set_id, attrs)
        return {'session_set': {}}

    return handle_session_data(handle, mode="a")


@api_bp.route('/get_session_sets', methods=['POST'])
@flask_login.login_required
def get_session_sets():
    """ Return all sets' name of this session. """
    def handle(session, set_id, **attrs):
        # set_id is not used here, but passed by default to the handle
        sets = session.data.get_sets(attrList=attrs)
        return {'session_sets': sets}

    return handle_session_data(handle, mode="r")


@api_bp.route('/add_session_item', methods=['POST'])
@flask_login.login_required
def add_session_item():
    """ Add a new item. """
    def handle(session, set_id, **attrs):
        itemId = attrs.pop("item_id")
        session.data.add_set_item(set_id, itemId, attrs)
        return {'item': {}}

    return handle_session_data(handle, mode="a")


@api_bp.route('/update_session_item', methods=['POST'])
@flask_login.login_required
def update_session_item():
    """ Update existing item. """
    def handle(session, set_id, **attrs):
        itemId = attrs.pop("item_id")
        session.data.update_set_item(set_id, int(itemId), attrs)
        return {'item': {}}

    return handle_session_data(handle, mode="a")


@api_bp.route('/get_session_data', methods=['POST'])
@flask_login.login_required
def get_session_data():
    """ Return some information related to session (e.g CTF values, etc). """
    def handle(session, set_id, **attrs):
        return DataContent(app).get_session_data(session)

    return handle_session_data(handle, mode="r")


# ---------------------------- INVOICE PERIODS --------------------------------

@api_bp.route('/get_invoice_periods', methods=['POST'])
@flask_login.login_required
def get_invoice_periods():
    return filter_request(app.dm.get_invoice_periods)


@api_bp.route('/create_invoice_period', methods=['POST'])
@flask_login.login_required
def create_invoice_period():
    return handle_invoice_period(app.dm.create_invoice_period)


@api_bp.route('/update_invoice_period', methods=['POST'])
@flask_login.login_required
def update_invoice_period():
    return handle_invoice_period(app.dm.update_invoice_period)


@api_bp.route('/delete_invoice_period', methods=['POST'])
@flask_login.login_required
def delete_invoice_period():
    return handle_invoice_period(app.dm.delete_invoice_period)


# ------------------------------ TRANSACTIONS ---------------------------------

@api_bp.route('/get_transactions', methods=['POST'])
@flask_login.login_required
def get_transactions():
    return filter_request(app.dm.get_transactions)


@api_bp.route('/create_transaction', methods=['POST'])
@flask_login.login_required
def create_transaction():
    return handle_transaction(app.dm.create_transaction)


@api_bp.route('/update_transaction', methods=['POST'])
@flask_login.login_required
def update_transaction():
    return handle_transaction(app.dm.update_transaction)


@api_bp.route('/delete_transaction', methods=['POST'])
@flask_login.login_required
def delete_transaction():
    return handle_transaction(app.dm.delete_transaction)


# ------------------------------ FORMS ---------------------------------

@api_bp.route('/get_forms', methods=['GET', 'POST'])
@flask_login.login_required
def get_forms():
    return send_json_data([f.json() for f in app.dm.get_forms()])


@api_bp.route('/create_form', methods=['POST'])
@flask_login.login_required
def create_form():
    return handle_form(app.dm.create_form)


@api_bp.route('/update_form', methods=['POST'])
@flask_login.login_required
def update_form():
    return handle_form(app.dm.update_form)


@api_bp.route('/delete_form', methods=['POST'])
@flask_login.login_required
def delete_form():
    return handle_form(app.dm.delete_form)


# ------------------------------ PROJECTS ---------------------------------

@api_bp.route('/get_projects', methods=['GET', 'POST'])
@flask_login.login_required
def get_projects():
    return send_json_data([p.json() for p in app.dm.get_projects()])


@api_bp.route('/create_project', methods=['POST'])
@flask_login.login_required
def create_project():
    return handle_project(app.dm.create_project)


@api_bp.route('/update_project', methods=['POST'])
@flask_login.login_required
def update_project():
    return handle_project(app.dm.update_project)


@api_bp.route('/delete_project', methods=['POST'])
@flask_login.login_required
def delete_project():
    return handle_project(app.dm.delete_project)


# ------------------------------ ENTRIES ---------------------------------

@api_bp.route('/get_entries', methods=['GET', 'POST'])
@flask_login.login_required
def get_entries():
    return send_json_data([p.json() for p in app.dm.get_entries()])


@api_bp.route('/create_entry', methods=['POST'])
@flask_login.login_required
def create_entry():
    def handle(**attrs):
        fix_dates(attrs, 'date')
        entry = app.dm.create_entry(**attrs)
        save_entry_files(entry, entry.extra['data'])
        entry = app.dm.update_entry(id=entry.id, extra=entry.extra)
        return entry.json()

    return _handle_item(handle, 'entry')


@api_bp.route('/update_entry', methods=['POST'])
@flask_login.login_required
def update_entry():
    def handle(**attrs):
        fix_dates(attrs, 'date', 'creation_date', 'last_update_date')
        entry = app.dm.get_entry_by(id=attrs['id'])
        old_files = set(app.dm.get_entry_files(entry))
        save_entry_files(entry, attrs['extra']['data'])
        entry = app.dm.update_entry(**attrs)
        new_files = set(app.dm.get_entry_files(entry))
        clean_files(old_files - new_files)
        return entry.json()

    return _handle_item(handle, 'entry')


@api_bp.route('/delete_entry', methods=['POST'])
@flask_login.login_required
def delete_entry():
    def handle(**attrs):
        entry = app.dm.get_entry_by(id=attrs['id'])
        old_files = set(app.dm.get_entry_files(entry))
        entry = app.dm.delete_entry(**attrs)
        clean_files(old_files)
        return entry.json()

    return _handle_item(handle, 'entry')


# ------------------------------ ENTRIES ---------------------------------

@api_bp.route('/get_pucks', methods=['GET', 'POST'])
@flask_login.login_required
def get_pucks():
    return send_json_data([p.json() for p in app.dm.get_pucks()])


@api_bp.route('/create_puck', methods=['POST'])
@flask_login.login_required
def create_puck():
    return handle_puck(app.dm.create_puck)


@api_bp.route('/update_puck', methods=['POST'])
@flask_login.login_required
def update_puck():
    return handle_puck(app.dm.update_puck)


@api_bp.route('/delete_puck', methods=['POST'])
@flask_login.login_required
def delete_puck():
    return handle_puck(app.dm.delete_puck)


# -------------------- UTILS functions ----------------------------------------

def filter_request(func):
    condition = request.json.get('condition', None)
    orderBy = request.json.get('orderBy', None)

    items = func(condition=condition, orderBy=orderBy,
                 asJson=True)

    if 'attrs' in request.json:
        attrs = request.json['attrs']

        def _filter(s):
            return {k: v for k, v in s.items()
                    if not attrs or k in attrs}
        items = [_filter(s) for s in items]

    return send_json_data(items)


def fix_dates(attrs, *date_keys):
    """ Convert the values from UTC string to datetime
    for some keys that might be present in the attrs dict. """
    for date_key in date_keys:
        if date_key in attrs:
            attrs[date_key] = datetime_from_isoformat(attrs[date_key])


def _handle_item(handle_func, result_key):
    try:
        if request.json:
            attrs = request.json['attrs']
        elif request.form:
            attrs = json.loads(request.form['attrs'])
        else:
            raise Exception("Expecting JSON or Form request.")
        result = handle_func(**attrs)
        return send_json_data({result_key: result})
    except Exception as e:
        print(e)
        import traceback
        traceback.print_exc()
        return send_error('ERROR from Server: %s' % e)


def handle_booking(result_key, booking_func, booking_transform=None):
    def handle(**attrs):
        dates = ['start', 'end']

        if attrs.get('repeat_value', 'no') != 'no':
            dates.append('repeat_stop')

        fix_dates(attrs, *dates)

        bt = booking_transform or app.dc.booking_to_event
        return [bt(b) for b in booking_func(**attrs)]

    return _handle_item(handle, result_key)


def handle_user(user_func):
    def handle(**attrs):
        fix_dates(attrs, 'created')
        return user_func(**attrs).json()

    return _handle_item(handle, 'user')


def handle_template(template_func):
    def handle(**attrs):
        return template_func(**attrs).json()

    return _handle_item(handle, 'template')


def handle_application(application_func):
    def handle(**attrs):
        return application_func(**attrs).json()

    return _handle_item(handle, 'application')


def handle_resource(resource_func):
    def handle(**attrs):
        r = resource_func(**attrs)

        for f in request.files:
            file = request.files[f]
            fn = file.filename
            if fn:
                # Clean up previous image files
                old_files = glob(app.dm.get_resource_image_path(r, '*'))
                clean_files(old_files)
                file.save(app.dm.get_resource_image_path(r, fn))
        return r.json()

    return _handle_item(handle, 'resource')


def handle_session(session_func):
    def handle(**attrs):
        fix_dates(attrs, 'start', 'end')
        return session_func(**attrs).json()

    return _handle_item(handle, 'session')


def handle_session_data(handle, mode="r"):
    attrs = request.json['attrs']
    session_id = attrs.pop("session_id")
    set_id = attrs.pop("set_id", None)
    tries = 0

    while tries < 3:
        try:
            session = app.dm.load_session(sessionId=session_id, mode=mode)
            result = handle(session, set_id, **attrs)
            session.data.close()
            break
        except OSError:
            print("Error with session data, sleeping 3 secs")
            time.sleep(3)
            result = {}
            tries += 1

    return send_json_data(result)


def handle_invoice_period(invoice_period_func):
    def handle(**attrs):
        fix_dates(attrs, 'start', 'end')
        return invoice_period_func(**attrs).json()

    return _handle_item(handle, 'invoice_period')


def handle_transaction(transaction_func):
    def handle(**attrs):
        fix_dates(attrs, 'date')
        return transaction_func(**attrs).json()

    return _handle_item(handle, 'transaction')


def handle_form(form_func):
    def handle(**attrs):
        return form_func(**attrs).json()

    return _handle_item(handle, 'form')


def handle_project(project_func):
    def handle(**attrs):
        fix_dates(attrs, 'date')
        return project_func(**attrs).json()

    return _handle_item(handle, 'project')


def save_entry_files(entry, data):
    """ Function used when creating or updating an Entry.
    It will handle the upload of new files.
    """
    for f in request.files:
        file = request.files[f]

        fn = file.filename
        if fn:
            data[f] = fn
            file.save(app.dm.get_entry_path(entry, fn))


def clean_files(paths):
    for p in paths:
        if os.path.exists(p):
            os.remove(p)


def handle_puck(puck_func):
    def handle(**attrs):
        return puck_func(**attrs).json()

    return _handle_item(handle, 'entry')


def create_item(name):
    def handle(**attrs):
        create_func = getattr(app.dm, 'create_%s' % name)
        item = create_func(**attrs)
        # For created items let's just return back the id
        return {'id': item.id}

    return _handle_item(handle, name)



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

import io
import numpy as np
import base64
import mrcfile

from PIL import Image, ImageEnhance, ImageOps


class Base64Converter:
    def __init__(self, **kwargs):
        self.max_size = kwargs.get('max_size', (512, 512))
        self.contrast_factor = kwargs.get('contrast_factor', None)
        self.scale = 1.0

    def from_pil(self, pil_img):
        """ Convert a PIL image into Base64. """
        if self.contrast_factor is not None:
            pil_img = ImageOps.autocontrast(pil_img, cutoff=self.contrast_factor)

        scale = 1.0
        w1, _ = pil_img.size
        if self.max_size is not None:
            pil_img.thumbnail(self.max_size)
            w2, _ = pil_img.size
            scale = w1 / w2

        self.scale = scale
        img_io = io.BytesIO()
        pil_img.save(img_io, format='PNG')

        return base64.b64encode(img_io.getvalue()).decode("utf-8")

    def from_path(self, path):
        """ Read the image path as a PIL image and encode it as base64.
        """
        try:
            img = Image.open(path)
            encoded = self.from_pil(img)
            img.close()
        except:
            encoded = ''

        return encoded

    def from_array(self, imageArray):
        # imean = imageArray.mean()
        # isd = imageArray.std()
        iMax = imageArray.max()  # min(imean + 10 * isd, imageArray.max())
        iMin = imageArray.min()  # max(imean - 10 * isd, imageArray.min())
        im255 = ((imageArray - iMin) / (iMax - iMin) * 255).astype(np.uint8)

        pil_img = Image.fromarray(im255)

        return self.from_pil(pil_img)

    def from_mrc(self, mrc_path):
        """ Convert real float32 mrc to base64.
        Convert to int8 first, then scale with Pillow.
        """
        mrc_img = mrcfile.open(mrc_path, permissive=True)

        if mrc_img.is_volume():
            imfloat = mrc_img.data[0, :, :]
        else:
            imfloat = mrc_img.data

        result = self.from_array(imfloat)
        mrc_img.close()

        return result

#
# def fn_to_blob(filename):
#     """ Read the image filename as a PIL image
#     and encode it as base64.
#     """
#     try:
#         with open(filename, 'rb') as img_f:
#             binary_data = img_f.read()  # read the image as python binary
#
#         return np.asarray(binary_data)
#     except:
#         return np.array(0)





# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *              Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [2]
# *
# * [1] SciLifeLab, Stockholm University
# * [2] MRC Laboratory of Molecular Biology (MRC-LMB)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

import unittest
import datetime as dt
from pprint import pprint

from emhub.data import (DataManager, ImageSessionData, H5SessionData,
                        PytablesSessionData, DataLog)
from emhub.data.imports.test import TestData
from emhub.utils import datetime_to_isoformat


class TestDataManager(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.dm = DataManager('/tmp/', cleanDb=True)
        # populate db with test data
        cls.td = TestData(cls.dm)

    def test_users(self):
        print("=" * 80, "\nTesting users...")
        users = self.dm.get_users()

        for u in users:
            pi = u.get_pi()

            if pi is not None:
                self.assertEqual(pi.id, u.get_pi().id)
            pi_str = pi.name if pi else 'None'

    def test_applications(self):
        print("=" * 80, "\nTesting applications...")
        applications = self.dm.get_applications()

        codes = [p.code for p in applications]

        self.assertEqual(codes, ['CEM00297', 'CEM00315', 'CEM00332', 'DBB00001',
                                 'CEM00345', 'CEM00346'])

        users = self.dm.get_users()
        uList = []
        piDict = {}

        for u in users:
            if u.is_pi:
                piDict[u.id] = [u]
                uList.append("PI: %s, applications: %s" % (u.name,
                                                           u.created_applications))
                uList.append("   Lab members:")
                for u2 in u.lab_members:
                    uList.append("     - %s" % u2.name)
            else:
                pi = u.get_pi()
                if pi is not None:
                    piDict[pi.id].append(u)

        for l in uList:
            print(l)

        # Check that all users in the same lab, have the same applications
        for pi_id, members in piDict.items():
            pi = members[0]
            pRef = pi.get_applications()

            # Check that the relationship pi-lab_members is working as expected
            print(">>> Checking PI: ", pi.name)
            members_ids = set(m.id for m in members[1:])
            lab_ids = set(u.id for u in pi.lab_members)
            print("   members: ", members_ids)
            print("   lab_ids: ", lab_ids)
            # self.assertEqual(members_ids, lab_ids)

            # Check applications for lab_members
            for u in members[1:]:
                self.assertEqual(pRef, u.get_applications())

    def test_bookings(self):
        # Retrieve all bookings that are either booking or downtime
        typeCond = "type='booking' OR type='downtime'"
        bookings = self.dm.get_bookings(condition=typeCond)
        self.assertEqual(len(bookings), 4)

        # Retrieve all bookings starting before or day 4
        fm = self.td.firstMonday + dt.timedelta(days=14)
        dateStr = datetime_to_isoformat(fm)
        startCond = "start<='%s' AND type='booking'" % dateStr
        bookings = self.dm.get_bookings(condition=startCond)
        self.assertEqual(len(bookings), 3)

    def test_count_booking_resources(self):
        print("=" * 80, "\nTesting counting booking resources...")

        def print_count(count):
            for a, count_dict in count.items():
                print("Application ID: ", a)
                for k, c in count_dict.items():
                    print("   %s: %s" % (k, c))

        applications = [a.id for a in self.dm.get_applications()]
        count_resources = self.dm.count_booking_resources(applications)
        print_count(count_resources)
        self.assertTrue(len(count_resources) > 0)

        count_tags = self.dm.count_booking_resources(applications,
                                                     resource_tags=['krios'])
        self.assertTrue(len(count_tags))
        print_count(count_tags)

    def test_resources(self):
        resources = self.dm.get_resources()

        microscopes = [r for r in resources if r.is_microscope]

        self.assertTrue(all(m.requires_slot for m in  microscopes))
        self.assertFalse(all(m.requires_slot for m in microscopes))


class TestSessionData(unittest.TestCase):
    def test_basic(self):
        setId = 1
        tsd = ImageSessionData(useBase64=True)
        FEW_ATTRS = ['id', 'ctfDefocus', 'location']

        mics = tsd.get_set_items(setId, attrList=FEW_ATTRS)

        self.assertEqual(len(mics), 20)

        firstMic = mics[0]
        self.assertEqual(list(firstMic.keys()), FEW_ATTRS)

        firstMicFull = tsd.get_set_item(setId, firstMic['id'],
                                        attrList=ImageSessionData.MIC_ALL_ATTRS)
        #pprint(firstMicFull)

        print("=" * 80, "\nTesting h5py session...")

        hsd = H5SessionData('/tmp/h5py.h5', 'w')
        micSetId = 'MicrographSet_%03d' % setId
        hsd.create_set(micSetId, {'label': 'Test set'})

        N = 10
        c = 0
        for i in range(N):
            for mic in mics:
                micFull = tsd.get_set_item(setId, mic['id'],
                                       attrList=ImageSessionData.MIC_ALL_ATTRS)
                c += 1
                print("\rWriting item %06d" % c, end='')
                hsd.add_set_item(micSetId, c, micFull)

        hsd.close()

        #hsd = H5SessionData('/tmp/h5py.h5', 'r')
        #mics2 = hsd.get_set_items(setId, attrList=ImageSessionData.MIC_ALL_ATTRS)
        #self.assertEqual(len(mics), len(mics2))



class TestPytablesSessionData(unittest.TestCase):
    def test_basic(self):
        setId = 1
        tsd = ImageSessionData(useBase64=True)
        FEW_ATTRS = ['id', 'ctfDefocus', 'location']

        mics = tsd.get_set_items(setId, attrList=FEW_ATTRS)

        print("=" * 80, "\nTesting pytables session...")
        psd = PytablesSessionData('/tmp/pytables.h5', 'w')
        psd.create_set(setId)

        for mic in mics:
            micData = tsd.get_item(setId, mic.id,
                                   dataAttrs=['micThumbData',
                                              'psdData',
                                              'shiftPlotData'])
            psd.add_item(setId, itemId=mic.id, **micData._asdict())

        # test updating item
        dic = {'location': '/new/location', 'psdData': 'newstr'}
        psd.update_item(setId, itemId=2, **dic)
        for mic in psd.get_items(setId,
                                 attrList=['id', 'location', 'ctfDefocus']):
            print(mic)

        psd.close()


class TestDataLog(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        pass

    def test(self):
        print("=" * 80, "\nTesting logs...")

        dbPath = '/tmp/emhub-logs.sqlite'
        dl  = DataLog(dbPath, cleanDb=True)

        logsData = [
            (1, 'data', 'create_test_user',
             ['Pepe Perez'], {'is_admin': False}),
            (1, 'error', 'deleting_file',
             [], {'error': 'File was locked', 'exception': True})
        ]

        for user_id, log_type, name, args, kwargs in logsData:
            dl.log(user_id, log_type, name, *args, **kwargs)

        dl.close()

        # Open again the logs and check
        dl = DataLog(dbPath)
        logs = dl.get_logs()
        self.assertEqual(2, len(logs))
        dl.close()



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *              Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [2]
# *
# * [1] SciLifeLab, Stockholm University
# * [2] MRC Laboratory of Molecular Biology (MRC-LMB)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

import os
from glob import glob


__version__ = '0.5.2'


def create_app(test_config=None):
    import flask
    import flask_login

    from . import utils
    from .blueprints import api_bp, images_bp, pages_bp
    from .utils import (datetime_to_isoformat,
                        pretty_date, pretty_datetime, pretty_quarter,
                        send_json_data, send_error)
    from .utils.mail import MailManager
    from .data.data_content import DataContent

    here = os.path.abspath(os.path.dirname(__file__))
    templates = [os.path.basename(f) for f in glob(os.path.join(here, 'templates', '*.html'))]

    # create and configure the app
    emhub_instance_path = os.environ.get('EMHUB_INSTANCE', None)

    app = flask.Flask(__name__,
                      instance_path=emhub_instance_path,
                      instance_relative_config=True)

    app.register_blueprint(api_bp, url_prefix='/api')
    app.register_blueprint(images_bp, url_prefix='/images')
    app.register_blueprint(pages_bp, url_prefix='/pages')

    app.config['SEND_FILE_MAX_AGE_DEFAULT'] = 0
    app.config['SECRET_KEY'] = 'dev'

    app.config["IMAGES"] = os.path.join(app.instance_path, 'images')
    app.config["USER_IMAGES"] = os.path.join(app.config["IMAGES"], 'user')
    app.config["ENTRY_FILES"] = os.path.join(app.instance_path, 'entry_files')
    app.config["RESOURCE_FILES"] = os.path.join(app.instance_path, 'resource_files')
    app.config["ALLOWED_IMAGE_EXTENSIONS"] = ["JPEG", "JPG", "PNG", "GIF"]
    app.config["SESSIONS"] = os.path.join(app.instance_path, 'sessions')
    app.config["PAGES"] = os.path.join(app.instance_path, 'pages')

    if test_config is None:
        # load the instance config, if it exists, when not testing
        app.config.from_pyfile('config.py', silent=True)
    else:
        # load the test config if passed in
        app.config.from_mapping(test_config)

    portalAPI = app.config.get('SLL_PORTAL_API', None)
    if portalAPI is not None:
        from .data.imports.scilifelab import PortalManager
        app.sll_pm = PortalManager(portalAPI, cache=False)

    # ensure the instance folder exists
    os.makedirs(app.config['USER_IMAGES'], exist_ok=True)
    os.makedirs(app.config['ENTRY_FILES'], exist_ok=True)
    os.makedirs(app.config['RESOURCE_FILES'], exist_ok=True)
    os.makedirs(app.config['SESSIONS'], exist_ok=True)
    os.makedirs(app.config['PAGES'], exist_ok=True)

    # Define some content_id list that does not requires login
    NO_LOGIN_CONTENT = ['users_list',
                        'user_reset_password',
                        'pages']

    @app.route('/main', methods=['GET', 'POST'])
    def main():
        if flask.request.method == 'GET':
            params = flask.request.args.to_dict()
            #content_id = flask.request.args.get('content_id', 'empty')
        else:
            params = flask.request.form.to_dict()
            #content_id = flask.request.form['content_id']

        content_id = params.pop('content_id', 'empty')
        kwargs = {'content_id': content_id,
                  'params': params
                  }

        if 'login_user' in params and app.is_devel:
            user_id = params['login_user']
            user = app.dm.get_user_by(id=user_id)
            if user is None:
                return send_error("Invalid user id: '%s'" % user_id)
            elif user != app.user:
                flask_login.logout_user()
                flask_login.login_user(user)

        if app.user.is_authenticated:
            if content_id == 'user_login':  # Redirects to Dashboard by default
                kwargs['content_id'] = 'dashboard'
            app.user.image = app.dc.user_profile_image(app.user)
        else:
            if content_id not in NO_LOGIN_CONTENT:
                kwargs = {'content_id': 'user_login',
                          'next_content': content_id,
                          'params': {}}

        kwargs['is_devel'] = app.is_devel
        kwargs['version'] = __version__
        kwargs['emhub_title'] = app.config.get('EMHUB_TITLE', '')
        kwargs['possible_owners'] = app.dc.get_pi_labs()
        kwargs['possible_operators'] = app.dc.get_possible_operators()
        kwargs.update(app.dc.get_resources_list())

        return flask.render_template('main.html', **kwargs)

    def _redirect(endpoint, **kwargs):
        return flask.redirect(flask.url_for(endpoint, **kwargs))

    @app.route('/', methods=['GET', 'POST'])
    @app.route('/index', methods=['GET', 'POST'])
    def index():
        return _redirect('pages.index', page_id='welcome')

    @app.route('/login', methods=['GET'])
    def login():
        """ This view will called when the user lands in the login page (GET)
        and also when login credentials are submitted (POST).
        """
        next_content = flask.request.args.get('next_content', 'empty')
        return _redirect('main', content_id=next_content)

    @app.route('/do_login', methods=['POST'])
    def do_login():
        """ This view will called as POST from the user login page. """
        username = flask.request.form['username']
        password = flask.request.form['password']
        next_content = flask.request.form.get('next_content', 'index')

        user = app.dm.get_user_by(username=username)
        if user is None or not user.check_password(password):
            flask.flash('Invalid username or password')
            return _redirect('login')

        flask_login.login_user(user)

        if next_content == 'user_login':
            next_content = 'dashboard'
        return _redirect('main', content_id=next_content)

    @app.route('/do_switch_login', methods=['POST'])
    @flask_login.login_required
    def do_switch_login():
        """ This view will called as POST from a currently logged admin user.
         It will allow admins to login as another users for troubleshooting.
        """
        if not app.user.is_admin:
            return send_error("Current user is not admin!")

        username = flask.request.json['username']

        user = app.dm.get_user_by(username=username)
        if user is None:
            return send_error("Invalid username: '%s'" % username)

        flask_login.logout_user()
        flask_login.login_user(user)

        return send_json_data({'user': user.id})

    @app.route('/logout', methods=['GET', 'POST'])
    def do_logout():
        flask_login.logout_user()
        return _redirect('index')

    @app.route('/reset_password', methods=['GET'])
    def reset_password():
        """ This view will called when the user lands in the login page (GET)
        and also when login credentials are submitted (POST).
        """
        return _redirect('main', content_id='user_reset_password')

    @app.route('/reset_password_request', methods=['POST'])
    def reset_password_request():
        """ This view will called as POST from the user login page. """
        email = flask.request.form['user-email']
        user = app.dm.get_user_by(email=email)

        if not email:
            msg = "ERROR: email can not be empty."
        elif user is None:
            msg = "ERROR: this email does not seen registered with any user."
        else:
            msg = ("Instructions about how to reset your password has been "
                   "sent to your email.")

            token = user.get_reset_password_token()
            app.dm.commit()  # store the user token

            def _render(fn):
                return flask.render_template(fn, user=user, token=token)

            app.mm.send_mail(
                [email],
                "emhub: CryoEM Reset your Password",
                _render('email/reset_password.txt'),
                html_body=_render('email/reset_password.html'))

        flask.flash(msg)
        return _redirect('reset_password')

    @app.route('/do_reset_password/<token>', methods=['GET', 'POST'])
    def do_reset_password(token):
        """ This view will called as POST from the user login page. """
        if app.user.is_authenticated:
            return _redirect('index')

        user = app.dm.User.verify_reset_password_token(token)
        app.dm.commit()  # store the user token

        if not user:
            flask.flash("ERROR: Invalid token for resetting password. ")
            return _redirect('main', content_id='user_reset_password')

        flask_login.login_user(user)

        return _redirect('main', content_id='user_form', user_id=user.id)

    @app.route('/get_content', methods=['GET', 'POST'])
    def get_content():
        if flask.request.method == 'GET':
            content_kwargs = flask.request.args.to_dict()
        else:
            content_kwargs = flask.request.form.to_dict()

        content_id = content_kwargs['content_id']

        if content_id in NO_LOGIN_CONTENT or app.user.is_authenticated:
            try:
                kwargs = app.dc.get(**content_kwargs)
            except Exception as e:
                import traceback
                tb = traceback.format_exc()
                error = {
                    'message': str(e),
                    'body': tb
                }
                return flask.render_template('error_dialog.html', error=error)
                # result = "<div><h1><span style='color: red'>Error</span> </br>%s</h1>" % e
                # result += "<pre>%s</pre></div>" % tb
                # return result
        else:
            kwargs = {'next_content': content_id}
            content_id = 'user_login'

        content_template = content_id + '.html'

        if content_template in templates:
            kwargs['is_devel'] = app.is_devel
            return flask.render_template(content_template, **kwargs)

        error = {
            "message": "Template '%s' not found." % content_template
        }
        return flask.render_template('error_dialog.html', error=error)


    @app.template_filter('basename')
    def basename(filename):
        return os.path.basename(filename)

    @app.template_filter('id_from_label')
    def id_from_label(label):
        return label.translate(label.maketrans("", "", "!#$%^&*()/\\ "))
        #return label.replace(' ', '_')

    @app.template_filter('range_params')
    def range_params(date_range):
        s = date_range[0].strftime("%Y/%m/%d")
        e = date_range[1].strftime("%Y/%m/%d")
        return "&start=%s&end=%s" % (s, e)

    def url_for_content(contentId, **kwargs):
        return flask.url_for('main', content_id=contentId, **kwargs)

    app.jinja_env.globals.update(url_for_content=url_for_content)
    app.jinja_env.add_extension('jinja2.ext.do')
    app.jinja_env.filters['reverse'] = basename
    app.jinja_env.filters['pretty_datetime'] = pretty_datetime
    app.jinja_env.filters['pretty_date'] = pretty_date
    app.jinja_env.filters['pretty_quarter'] = pretty_quarter

    from emhub.data.data_manager import DataManager
    app.user = flask_login.current_user
    app.dm = DataManager(app.instance_path, user=app.user)
    app.dc = DataContent(app)
    app.is_devel = (os.environ.get('FLASK_ENV', None) == 'development')
    app.version = __version__

    login_manager = flask_login.LoginManager()
    #login_manager.login_view = 'login'
    login_manager.init_app(app)

    app.mm = MailManager(app)

    from flaskext.markdown import Markdown
    Markdown(app)

    @login_manager.user_loader
    def load_user(user_id):
        return app.dm.get_user_by(id=int(user_id))

    @app.teardown_appcontext
    def shutdown_session(exception=None):
        app.dm.close()

    return app



#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

""" 
This script will monitors the pre-processing progress in Scipion and notify to EMhub.
"""

import sys, os
import time
import datetime as dt
import argparse
from pprint import pprint
from contextlib import contextmanager

import mrcfile
from pyworkflow.project import Manager
import pyworkflow.utils as pwutils
from pwem.objects import SetOfCTF


from emhub.client import open_client
from emhub.utils.image import Base64Converter


def usage(error):
    print("ERROR: %s" % error)
    get_parser().print_usage()
    sys.exit(1)


class SetMonitor:
    """ Monitor when there are changes to a given set. """
    def __init__(self, inputSet):
        self.__class = inputSet.getClass()
        self.__filename = inputSet.getFileName()
        self.__lastCheck = None
        self.count = 0

    @contextmanager
    def open_set(self):
        modified = pwutils.getFileLastModificationDate(self.__filename)

        if self.__lastCheck is None or modified > self.__lastCheck:
            setObj = self.__class(filename=self.__filename)
            setObj.loadAllProperties()
            yield setObj
            setObj.close()
        else:
            yield None

        self.__lastCheck = modified

    def update_count(self):
        """ Return True if there were new items. """
        old_count = self.count
        with self.open_set() as setObj:
            if setObj is not None:
                self.count = setObj.getSize()

        return self.count > old_count


def get_parser():
    """ Return the argparse parser, so we can get the arguments """

    parser = argparse.ArgumentParser()
    add = parser.add_argument  # shortcut

    g = parser.add_mutually_exclusive_group()
    g.add_argument('--list', action='store_true',
                   help="List existing sessions in the server.")
    g.add_argument('--update', action='store_true',
                   help="Update an existing session")

    add('-p', '--project', metavar='PROJECT_NAME',
        help='Project name.')

    add('--session_id', type=int,
        help='Session Id')

    add('--prot_ctf', type=int, default=0,
        help="Id of the CTF protocol to monitor. ")

    add('--prot_picking', type=int, default=0,
        help="Id of the particle picking protocol to monitor. ")

    add('--prot_2d', type=int, default=0,
        help='Id of the classify 2d protocol.')

    add('--clear', action='store_true',
        help="Clear existing data associated with the session.")

    return parser


class ProjectSession:
    def __init__(self, projName, sessionId, protIds):
        self._sessionId = sessionId
        manager = Manager()

        if not manager.hasProject(projName):
            usage("No project found with name '%s'" % pwutils.red(projName))

        if not 'ctf' in protIds:
            usage("Please provide a valid CTF-protocol ID.")

        self._project = manager.loadProject(projName)
        self._protocols = {key: self._load_protocol(protId)
                           for key, protId in protIds.items()}

    def _get_path(self, *paths):
        return os.path.join(self._project.path, *paths)

    def _load_protocol(self, protId):
        try:
            return self._project.getProtocol(protId)
        except:
            usage("No protocol found with ID '%s'" % pwutils.red(protId))

    def _update_mics_ctfs(self, dc):
        protCtf = self._protocols['ctf']
        outputCTF = getattr(protCtf, 'outputCTF', None)

        micSet = protCtf.inputMicrographs.get()
        self._micSetId = 'Micrographs_%06d' % micSet.getObjId()
        acq = micSet.getAcquisition()

        attrs = {
            'session_id': self._sessionId,
            'set_id': self._micSetId
        }

        stats = {
            'numOfCls2D': 0,
            'numOfCtfs': outputCTF.getSize(),
            'numOfMics': micSet.getSize(),
            'numOfMovies': 0,
            'numOfPtcls': 0,
            'ptclSizeMax': 0,
            'ptclSizeMin': 0
        }

        micMonitor = SetMonitor(micSet)
        # ctfMonitor = SetMonitor(outputCTF)

        found_new_mics = False
        ctfSet = SetOfCTF(filename=outputCTF.getFileName())
        ctfSet.loadAllProperties()
        lastId = 0

        new_stats = {}
        micBase64 = Base64Converter()
        psdBase64 = Base64Converter()

        for ctf in ctfSet.iterItems(where="id>%s" % lastId):
            u, v, a = ctf.getDefocus()
            lastId = ctfId = ctf.getObjId()
            mic = ctf.getMicrograph()
            pixelSize = mic.getSamplingRate()

            attrs.update({
                'item_id': ctfId,
                'ctfDefocus': (u + v) * 0.5,
                'ctfDefocusU': u,
                'ctfDefocusV': v,
                'ctfDefocusAngle': a,
                'ctfResolution': ctf.getResolution(),
                'ctfFit': ctf.getFitQuality(),
                'location': mic.getFileName(),
                'ctfFitData': '',
                'shiftPlotData': '',
                'pixelSize': pixelSize
            })

            print("Adding item %06d" % ctfId)
            psdPath = os.path.join(self._project.path, ctf.getPsdFile())

            if os.path.exists(psdPath):
                print("  PSD: ", psdPath)
                attrs['psdData'] = psdBase64.from_mrc(psdPath)

            micPath = os.path.join(self._project.path, ctf.getMicrograph().getFileName())
            if os.path.exists(micPath):
                print("  MIC: ", micPath)
                attrs['micThumbData'] = micBase64.from_mrc(micPath)
                attrs['micThumbPixelSize'] = pixelSize * micBase64.scale

            dc.add_session_item(attrs)

        new_stats['numOfCtfs'] = ctfSet.getSize()

        # Check if there are new micrographs
        if micMonitor.update_count():
            new_stats['numOfMics'] = micMonitor.count

        if new_stats:
            stats.update(new_stats)
            print("Updating session stats: ")
            print("   Mics: ", micMonitor.count)
            print("   CTFs: ", ctfSet.getSize())

            dc.update_session({'id': self._sessionId, 'stats': stats})
        else:
            time.sleep(10)

        ctfSet.close()

        print("lastId: ", lastId)
            #
            # if ctfSet.isStreamClosed():
            #     with open_client() as dc:
            #         dc.update_session({'id': self._sessionId, 'status': 'finished'})
            #     break

    def _update_coords(self, dc):
        protCtf = self._protocols['ctf']
        outputCTF = getattr(protCtf, 'outputCTF', None)
        micSet = protCtf.inputMicrographs.get()

        protPicking = self._protocols.get('picking', None)

        if protPicking is None:
            return

        coordsSet = getattr(protPicking, 'outputCoordinates', None)
        attrs = {
            'session_id': self._sessionId,
            'set_id': 'Micrographs_%06d' % micSet.getObjId()
        }
        coordList = []
        lastMicId = None

        for coord in coordsSet.iterItems(orderBy='_micId', direction='ASC'):
            micId = coord.getMicId()
            if micId != lastMicId:
                # Update the coordinates information if necessary
                if lastMicId is not None:
                    attrs['item_id'] = lastMicId
                    attrs['coordinates'] = coordList

                    dc.update_session_item(attrs)
                lastMicId = micId
                coordList = []
            coordList.append(coord.getPosition())

    def _update_classes(self, dc):
        prot2D = self._protocols.get('2d', None)

        if prot2D is None:
            return

        sets = dc.get_session_sets({'session_id': self._sessionId})
        outputClasses = getattr(prot2D, 'outputClasses', None)
        outputClasses.printAll()
        setId = 'Class2D_%06d' % outputClasses.getObjId()
        attrs = {
            'session_id': self._sessionId,
            'set_id': setId,
            'size': outputClasses.getSize()
        }
        if any(s['id'] == setId for s in sets):
            set_func = dc.update_session_set
            item_func = dc.update_session_item
            label = 'Updating'
        else:
            set_func = dc.create_session_set
            item_func = dc.add_session_item
            label = 'Creating'

        print("- %s set %s" % (label, setId))
        set_func(attrs)

        fn = outputClasses.getFirstItem().getRepresentative().getFileName()
        mrc_stack = mrcfile.open(fn, permissive=True)
        stackBase64 = Base64Converter(max_size=None)

        for class2d in outputClasses:
            rep = class2d.getRepresentative()
            rep.setFileName(self._get_path(rep.getFileName()))
            i = rep.getIndex() - 1
            attrs.update({
                'item_id': class2d.getObjId(),
                'size': class2d.getSize(),
                'average': stackBase64.from_array(mrc_stack.data[i, :, :])
            })
            for i in range(3):  # try 3 times
                try:
                    item_func(attrs)
                    break
                except Exception as e:
                    print("dc.add_session_item:: Error: %s" % e)
                    print("                      Trying again in 3 seconds.")
                    time.sleep(3)

    def run(self):
        with open_client() as dc:
            sessionDict = dc.get_session(self._sessionId)
            self._update_mics_ctfs(dc)
            self._update_coords(dc)
            self._update_classes(dc)


def main():
    args = get_parser().parse_args()

    if args.clear:
        with open_client() as dc:
            print("Clearing session data: ")
            dc.request('clear_session_data', jsonData={'attrs': {'id': args.session_id}})

    if args.list:
        with open_client() as dc:
            r = dc.request('get_sessions', jsonData={})

            for session in r.json():
                pprint(session)

    elif args.prot_ctf:

        protocols = {'ctf': args.prot_ctf}
        if args.prot_picking:
            protocols['picking'] = args.prot_picking
        if args.prot_2d:
            protocols['2d'] = args.prot_2d

        ProjectSession(args.project, args.session_id, protocols).run()

    else:
        print("Please provide some arguments")
        sys.exit(1)


if __name__ == '__main__':
    main()








# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *              Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [2]
# *
# * [1] SciLifeLab, Stockholm University
# * [2] MRC Laboratory of Molecular Biology (MRC-LMB)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

import os
import datetime as dt
import json
from collections import defaultdict

import flask
import flask_login

from emhub.utils import (pretty_datetime, datetime_to_isoformat, pretty_date,
                         datetime_from_isoformat, get_quarter, pretty_quarter,
                         image)


class DataContent:
    """ This class acts as an intermediary between the DataManager and
    the Flask application.

    Here information is retrieved from the DataManager (dealing with stored
    data structure) and prepare the "content" for required views.
    """
    def __init__(self, app):
        """ Create a new content for the given Flask application. """
        self.app = app

    def _dateStr(self, datetime):
        return

    def get(self, **kwargs):
        content_id = kwargs['content_id']
        get_func_name = 'get_%s' % content_id.replace('-', '_')  # FIXME
        dataDict = {}  # self.get_resources_list()
        get_func = getattr(self, get_func_name, None)
        if get_func is not None:
            dataDict.update(get_func(**kwargs))
        return dataDict

    def get_dashboard(self, **kwargs):
        dataDict = self.get_resources_list()
        user = self.app.user  # shortcut
        resource_bookings = {}

        # Provide upcoming bookings sorted by proximity
        bookings = [('Today', []),
                    ('Next 7 days', []),
                    ('Next 30 days', [])]


        now  = self.app.dm.now()
        next7 = now + dt.timedelta(days=7)
        next30 = now + dt.timedelta(days=30)

        for b in self.app.dm.get_bookings(orderBy='start'):
            if not user.is_manager and not user.same_pi(b.owner):
                continue
            bDict = {'owner': b.owner.name,
                     'resource': b.resource.name,
                     'start': pretty_datetime(b.start),
                     'end': pretty_datetime(b.end),
                     }
            if b.start <= now <= b.end:
                i = 0
            elif now <= b.start <= next7:
                i = 1
            elif now <= b.start <= next30:
                i = 2
            else:
                i = -1

            if i >= 0:
                bookings[i][1].append(bDict)
                r = b.resource
                if i == 0 and r.is_microscope and 'solna' in r.tags:  # Today's bookings
                    # FIXME: If there is already a session, also return its id
                    resource_bookings[r.id] = b

        dataDict.update({'bookings': bookings,
                         'lab_members': self.get_lab_members(user),
                         'resource_bookings': resource_bookings})
        return dataDict

    def get_lab_members(self, user):
        if user.is_staff:
            return [u.json() for u in self._get_facility_staff(user.staff_unit)]

        pi = user.get_pi()
        if pi is None:
            return []
        members = [u.json() for u in pi.get_lab_members()]
        members.insert(0, pi.json())
        return members

    def get_sessions_overview(self, **kwargs):
        sessions = self.app.dm.get_sessions(condition=self._get_display_condition(),
                                            orderBy='resource_id')
        return {'sessions': sessions}

    def get_session_data(self, session):
        micSetId = None
        sets = session.data.get_sets()
        for s in sets:
            if not 'id' in s:
                print("ERROR, set without id: ", s)
                continue
            if s['id'].startswith('Micrographs'):
                micSetId = s['id']

        if micSetId is None:
            raise Exception("Not micrograph set found in '%s'"
                            % session.data_path)

        attrList = ['location', 'ctfDefocus', 'ctfResolution']
        mics = session.data.get_set_items(micSetId, attrList=attrList)

        def _get_hist(label, inputList):
            import numpy as np
            hist, bins = np.histogram(inputList)
            return {
                'label': label,
                'data': [int(h) for h in hist],
                'bins': ['%0.1f' % b for b in bins],
            }

        defocusList = [m['ctfDefocus'] for m in mics]
        resolutionList = [m['ctfResolution'] for m in mics]
        stats = session.stats

        # Load classes
        classesSet = [s for s in sets if s['id'].startswith('Class2D')]
        if classesSet:
            class2DSetId = classesSet[-1]['id']
            classes2d = session.data.get_set_items(class2DSetId, attrList=['size', 'average'])
            classes2d.sort(key=lambda c: c['size'], reverse=True)
        else:
            classes2d = []

        session.data.close()

        return {
            'defocus_plot': ['Defocus'] + defocusList,
            'ctf_defocus_hist': _get_hist('CTF Defocus', defocusList),
            'resolution_plot': ['Resolution'] + resolutionList,
            'ctf_resolution_hist': _get_hist('CTF Resolution', resolutionList),
            'session': session.json(),
            'counters': {
                'imported': stats['numOfMics'],
                'aligned': stats['numOfMics'],
                'ctf': stats['numOfCtfs'],
                'picked': stats['numOfPtcls']
            },
            'classes2d': classes2d
        }

    def get_session_live(self, **kwargs):
        session_id = kwargs['session_id']
        session = self.app.dm.load_session(session_id)
        return self.get_session_data(session)

    def get_session_details(self, **kwargs):
        session_id = kwargs['session_id']
        session = self.app.dm.get_session_by(id=session_id)
        if session.booking:
            a = session.booking.application
            if not (a is None or a.allows_access(self.app.user)):
                raise Exception("You do not have access to this session information. ")

        days = self.app.dm.get_session_data_deletion(session.name[:3])
        td = (session.start + dt.timedelta(days=days)) - self.app.dm.now()
        errors = []
        # TODO: We might check other type of errors in the future
        status_info = session.extra.get('status_info', '')
        if status_info.lower().startswith('error:'):
            errors.append(status_info)

        return {
            'session': session,
            'deletion_days': td.days,
            'errors': errors
        }

    def get_sessions_list(self, **kwargs):
        dm = self.app.dm  # shortcut
        all_sessions = dm.get_sessions()
        sessions = []
        bookingDict = {}

        for s in all_sessions:
            if s.booking:
                a = s.booking.application
                if a is None or a.allows_access(self.app.user):
                    sessions.append(s)
                    b = self.booking_to_event(s.booking,
                                              prettyDate=True, piApp=True)
                    bookingDict[s.booking.id] = b

                    if not os.path.exists(dm.get_session_data_path(s)):
                        s.data_path = ''

        return {
            'sessions': sessions,
            'bookingDict': bookingDict,
            'possible_owners': self.get_pi_labs(),
            'possible_operators': self.get_possible_operators(),
        }

    def get_users_list(self, **kwargs):
        users = self.app.dm.get_users()
        for u in users:
            u.image = self.user_profile_image(u)
            u.project_codes = [p.code for p in u.get_applications()]

        return {'users': users}

    def get_users_groups_cards(self, **kwargs):
        # Retrieve all pi labs that belong to a given application

        appCode = kwargs.get('application', '').upper()

        def _userjson(u):
            return {'id': u.id, 'name': u.name}

        # Group users by PI
        labs = []

        if appCode:
            piList = [u for u in self.app.dm.get_users()
                      if u.is_pi and u.has_application(appCode)]

            for u in piList:
                if u.is_pi:
                    lab = [_userjson(u)] + [_userjson(u2) for u2 in u.get_lab_members()]
                    labs.append(lab)

            labs = sorted(labs, key=lambda lab: len(lab), reverse=True)

        apps = sorted(self.app.dm.get_visible_applications(),
                      key=lambda a: len(a.users), reverse=True)
        applications = [a.json() for a in apps if a.is_active]

        # if user.is_staff:
        #     labs.append([_userjson(u) for u in self._get_facility_staff(user.staff_unit)])

        return {
            'labs': labs,
            'applications': applications,
            'application_code': appCode,
        }

    def get_users_groups(self, **kwargs):
        return self.get_users_groups_cards(**kwargs)

    def get_user_form(self, **kwargs):
        user = self.app.dm.get_user_by(id=kwargs['user_id'])
        user.image = self.user_profile_image(user)
        data = {'user': user}

        pi_label = None

        # Allow to change pi if it is a manager logged
        if user.is_pi:
            pi_label = 'Self'
        elif user.is_manager:
            pi_label = 'None'
        elif self.app.user.is_manager:
            data['possible_pis'] = [
                {'id': u.id, 'name': u.name}
                for u in self.app.dm.get_users() if u.is_pi
            ]
        else:
            pi = user.pi
            pi_label = 'Unknown' if pi is None else pi.name

        data['pi_label'] = pi_label
        data['user_statuses'] = self.app.dm.User.STATUSES

        return data

    def get_register_user_form(self, **kwargs):
        dm = self.app.dm  # shortcut

        return {
            'possible_pis': [{'id': u.id, 'name': u.name}
                             for u in dm.get_users() if u.is_pi],
            'pi_label': None,
            'roles': dm.User.ROLES
        }

    def get_resources_list(self, **kwargs):
        user = self.app.user
        if not user.is_authenticated:
            return  {'resources': []}

        def _image(r):
            fn = self.app.dm.get_resource_image_path(r)
            if os.path.exists(fn):
                base64 = image.Base64Converter(max_size=(1024, 1024))
                return 'data:image/%s;base64, ' + base64.from_path(fn)
            else:
                return flask.url_for('images.static', filename=r.image)

        resource_list = [
            {'id': r.id,
             'name': r.name,
             'status': r.status,
             'tags': r.tags,
             'requires_slot': r.requires_slot,
             'latest_cancellation': r.latest_cancellation,
             'color': r.color,
             'image': _image(r),
             'user_can_book': user.can_book_resource(r),
             'is_microscope': r.is_microscope,
             'min_booking': r.min_booking,
             'max_booking': r.max_booking,
             'daily_cost': r.daily_cost
             }
            for r in self.app.dm.get_resources()
        ]
        return {'resources': resource_list}

    def get_resource_form(self, **kwargs):
        copy_resource = json.loads(kwargs.pop('copy_resource', 'false'))

        r = self.app.dm.get_resource_by(id=kwargs['resource_id'])

        if r is None:
            r = self.app.dm.Resource(
                name='',
                status='inactive',
                tags='',
                image='',
                color='rgba(256, 256, 256, 1.0)',
                extra={})

        if copy_resource:
            r.id = None
            r.name = 'COPY ' + r.name

        params = []

        def _add(attr, label, **kwargs):
            p = {'id': attr,
                 'label': label,
                 'value': getattr(r, attr)
                 }
            p.update(kwargs)
            params.append(p)

        _add('name', 'Name')
        _add('status', 'Status',
             enum={'display': 'combo',
                   'choices': ['active', 'inactive']})
        _add('tags', 'Tags')
        _add('image', 'Icon image', type='file_image')
        _add('color', 'Color')
        _add('latest_cancellation', 'Latest cancellation (h)')
        _add('min_booking', 'Minimum booking time (h)')
        _add('max_booking', 'Maximum booking time (h)')
        _add('daily_cost', 'Daily cost')
        _add('requires_slot', 'Requires Slot', type='bool')
        _add('requires_application', 'Requires Application', type='bool')

        return {
            'resource': r, 'params': params
        }

    def get_booking_calendar(self, **kwargs):
        dm = self.app.dm  # shortcut
        dataDict = self.get_resources_list()
        dataDict['bookings'] = [self.booking_to_event(b)
                                for b in dm.get_bookings()
                                if b.resource is not None]
        dataDict['current_user_json'] = flask_login.current_user.json()
        dataDict['applications'] = [{'id': a.id,
                                     'code': a.code,
                                     'alias': a.alias}
                                    for a in dm.get_applications()
                                    if a.is_active]

        dataDict['possible_owners'] = self.get_pi_labs()
        dataDict['possible_operators'] = self.get_possible_operators()

        dataDict['resource_id'] = kwargs.get('resource_id', None)
        return dataDict

    def get_applications(self, **kwargs):
        dataDict = self.get_raw_applications_list()
        dataDict['template_statuses'] = ['preparation', 'active', 'closed']
        dataDict['template_selected_status'] = kwargs.get('template_selected_status', 'active')
        dataDict['templates'] = [{'id': t.id,
                                  'title': t.title,
                                  'description': t.description,
                                  'status': t.status,
                                  'iuid': t.extra.get('portal_iuid', 'no'),
                                  'code_prefix': t.code_prefix
                                  }
                                 for t in self.app.dm.get_templates()]

        return dataDict

    def get_application_form(self, **kwargs):
        dm = self.app.dm  # shortcut

        if 'application_id' in kwargs:
            app = dm.get_application_by(id=kwargs['application_id'])
        else:  # New Application
            template = dm.get_template_by(id=kwargs['template_id'])
            if not template.code_prefix:
                raise Exception("It is not possible to create Applications from this template. \n"
                                "Maybe they need to be imported from the Portal. ")

            max_code = 0
            for a in dm.get_applications():
                code = a.code
                if code.startswith(template.code_prefix):
                    try:
                        max_code = max(max_code, int(code[3:]))
                    except:
                        pass

            app = dm.Application(code='%s%05d' % (template.code_prefix.upper(), max_code + 1),
                                 title='', alias='', description='',
                                 creator=self.app.user,
                                 resource_allocation=dm.Application.DEFAULT_ALLOCATION,
                                 extra={})

        # Microscopes info to setup some permissions on the Application form
        mics = [{'id': r.id,
                 'name': r.name,
                 'noslot': app.no_slot(r.id),
                 } for r in dm.get_resources() if r.is_microscope]

        # Check which PIs are in the application
        in_app = set(pi.id for pi in app.pi_list)

        return {'application': app,
                'application_statuses': dm.Application.STATUSES,
                'template_id': kwargs.get('template_id', None),
                'microscopes': mics,
                'pi_list': [{'id': u.id,
                             'name': u.name,
                             'email': u.email,
                             'in_app': u.id in in_app,
                             'status': 'creator' if u.id == app.creator.id else ''
                             }
                            for u in dm.get_users() if u.is_pi],
                'users': [u for u in dm.get_users() if u.is_staff]
                }

    def set_form_values(self, form, values):
        """ Load values to form parameters based on the ids.
        """
        definition = form.definition

        def set_value(p):
            if 'id' not in p:
                return
            p['value'] = values.get(p['id'], p.get('default', ''))

        if 'params' in definition:
            for p in definition['params']:
                set_value(p)
        else:
            for section in definition['sections']:
                for p in section['params']:
                    set_value(p)

    def get_experiment_form(self, **kwargs):
        data = self.get_dynamic_form_modal(**kwargs)
        data.update(self.get_grids_storage())
        return data

    def get_dynamic_form_modal(self, **kwargs):
        form_id = int(kwargs.get('form_id', 1))
        form_values_str = kwargs.get('form_values', None) or '{}'
        form_values = json.loads(form_values_str)

        form = self.app.dm.get_form_by(id=form_id)

        if form is None:
            raise Exception("Invalid form id: %s" % form_id)

        self.set_form_values(form, form_values)

        return {'form': form}

    def get_logs(self, **kwargs):
        dm = self.app.dm
        logs = int(kwargs.get('n', 100))
        logs = dm.get_logs()[-logs:  ]
        for log in logs:
            log.user = dm.get_user_by(id=log.user_id)

        return  {'logs': logs}

    def get_pages(self, **kwargs):
        page_id = kwargs['page_id']
        page_path = os.path.join(self.app.config['PAGES'], '%s.md' % page_id)

        return {
            'page_id': page_id,
            'page': 'pages/%s.md' % page_id
        }

    def get_portal_users_list(self, **kwargs):
        dm = self.app.dm
        do_import = 'import' in kwargs
        imported = []
        failed = []

        if do_import:
            users = self._get_users_from_portal(status='ready')
            for u in users:
                roles = ['user', 'pi'] if u['pi'] else ['user']
                pi_id = None if u['pi'] else u['pi_user'].id

                try:
                    user = dm.create_user(
                        username=u['email'],
                        email=u['email'],
                        phone='',
                        password=os.urandom(24).hex(),
                        name="%(first_name)s %(last_name)s" % u,
                        roles=roles,
                        pi_id=pi_id,
                        status='active'
                    )
                    imported.append(u)
                    self.app.mm.send_mail(
                        [user.email],
                        "emhub: New account imported",
                        flask.render_template('email/account_created.txt',
                                              user=user))
                except Exception as e:
                    u['error'] = str(e)
                    failed.append(u)
            status = None
        else:
            status = kwargs.get('status', None)

        users = self._get_users_from_portal(status)

        return {'portal_users': users,
                'status': status,
                'do_import': do_import,
                'users_imported': imported,
                'users_failed': failed
                }

    def get_portal_import_application(self, **kwargs):
        # Date since the created orders in the portal will be considered
        sinceArg = kwargs.get('since', None)

        if sinceArg:
            since = datetime_from_isoformat(sinceArg)
        else:
            since = self.app.dm.now() - dt.timedelta(days=183)  # 6 months

        result = {'since': since}

        app_code = None
        if app_code is not None:
            try:
                app_code = app_code.upper()
                app = self._import_order_from_portal(app_code)
                result['order'] = app
            except Exception as e:
                result['errors'] = [str(e)]

        else:
            ordersJson = self.app.sll_pm.fetchOrdersJson()

            def _filter(o):
                s = o['status']
                code = o['identifier'].upper()
                app = self.app.dm.get_application_by(code=code)
                created = datetime_from_isoformat(o['created'])

                return ((s == 'accepted' or s == 'processing') and
                        app is None and created >= since)
                        #created.year == self.app.dm.now().year)

            result['orders'] = [o for o in ordersJson if _filter(o)]

        return result

    def get_reports_time_distribution(self, **kwargs):

        bookings, range_dict = self.get_booking_in_range(kwargs)

        from emhub.reports import get_booking_counters
        counters, cem_counters = get_booking_counters(bookings)

        details_key = kwargs.get('details', None) or 'Reminder'

        def _group(bookings):
            pi_bookings = {}
            for b in bookings:
                pi_id = b['pi_id']
                if pi_id not in pi_bookings:
                    pi_bookings[pi_id] = [[pi_id, b['pi_name'], 0, 0]]

                pi_list = pi_bookings[pi_id]
                pi_list.append(b)
                pi_list[0][2] += b['days']
                pi_list[0][3] += b['total_cost']

            return list(pi_bookings.values())

        app_dict = {a.code: a.alias for a in self.app.dm.get_applications()}
        if details_key.startswith('CEM') and len(details_key) > 3:
            alias = app_dict.get(details_key, None)
            details_title = details_key + (' (%s)' % alias if alias else '')
            details_bookings = cem_counters[details_key].bookings
        else:
            details_title = {
                'Reminder': "Reminder: Uncategorized Bookings (Review and Update)",
                'DBB': "DBB Bookings",
                'Downtime': "Downtime",
                'Maintenance': "Maintenance",
                'Development': "Development"
            }[details_key]
            details_bookings = counters[details_key].bookings

        d = {
            'overall': counters,
            'cem': cem_counters,
            'possible_owners': self.get_pi_labs(),
            'possible_operators': self.get_possible_operators(),
            'app_dict': app_dict,
            'details_bookings': details_bookings,
            'details_title': details_title,
            'details_key': details_key,
            'details_groups': [],
        }

        d.update(range_dict)

        if details_key.startswith('CEM') or details_key.startswith('DBB'):
            d['details_groups'] = _group(details_bookings)

        return d

    def get_reports_invoices(self, **kwargs):
        bookings, range_dict = self.get_booking_in_range(kwargs, asJson=False)

        portal_users = {
            pu['email']: pu for pu in self.app.sll_pm.fetchAccountsJson()
            if pu['pi']
        }

        def create_pi_info_dict(a):
            return {pi.id: {
                'pi_name': pi.name,
                'pi_email': pi.email,
                'bookings': [],
                'sum_cost': 0,
                'sum_days': 0,
                } for pi in a.pi_list
            }

        def update_pi_info(pi_info, b):
            pi_info['bookings'].append(b)
            pi_info['sum_cost'] += b.total_cost
            pi_info['sum_days'] += b.days

        # Create a dictionary where pi/bookings are grouped by Application
        # and another one grouped by pi
        apps_dict = {}
        pi_dict = {}

        for a in self.app.dm.get_applications():
            apps_dict[a.code] = create_pi_info_dict(a)
            pi_dict.update(create_pi_info_dict(a))

        for b in bookings:
            if b.application is None:
                continue

            app_id = b.application.code

            if  app_id not in apps_dict:
                print(">>> Missing app: ", app_id)
                continue

            pi = b.owner.get_pi()

            if pi is None:
                print(">>>  None pi, user: ", b.owner.name)
                continue

            try:
                update_pi_info(apps_dict[app_id][pi.id], b)
                update_pi_info(pi_dict[pi.id], b)

                if b.total_cost == 0:
                    print(">>> 0 cost booking1!!!")
                    print(b.json())

            except KeyError:
                print("Got KeyError, app_id: %s, pi_id: %s"
                      % (app_id, pi.id))

        result = {
            'apps_dict':  apps_dict,
            'pi_dict': pi_dict,
            'portal_users': portal_users,
            'group': int(kwargs.get('group', 1))
        }
        result.update(range_dict)

        return result

    def __get_period(self, kwargs):
        """ Helper function to update period in kwargs if not present. """
        dm = self.app.dm  # shortcut
        period = None
        period_id = kwargs.get('period', None)

        if period_id is None:
            for p in dm.get_invoice_periods():
                if p.status == 'active':
                    period = p
        else:
            period = dm.get_invoice_period_by(id=int(period_id))

        kwargs['start'] = pretty_date(period.start)
        kwargs['end'] = pretty_date(period.end)

        return period

    def __get_pi_user(self, kwargs):
        """ Helper function to get pi user. """
        u = self.app.user
        pi_id_value = kwargs.get('pi_id', u.id)
        bag_visible = kwargs.get('bag_visible', True)

        try:
            pi_id = int(pi_id_value)

        except Exception:
            raise Exception("Provide a valid integer id.")

        pi_user = self.app.dm.get_user_by(id=pi_id)
        if pi_user is None:
            raise Exception("Invalid user id: %s" % pi_id)

        def _has_access():
            if u.is_manager:
                return True

            if not u.is_pi:
                return False

            # When bag_visible is False, only the own PI can bee seen
            if not bag_visible and u.id != pi_user.id:
                return False

            pi_apps = set(a.id for a in pi_user.get_applications())
            u_apps = set(a.id for a in u.get_applications())

            return bool(pi_apps.intersection(u_apps))

        if not _has_access():
            raise Exception("You do not have access to this information. u.is_pi: %s" % u.is_pi)

        return pi_user

    def get_reports_invoices_lab(self, **kwargs):
        period = self.__get_period(kwargs)
        pi_user = self.__get_pi_user(kwargs)
        app = pi_user.get_applications()[-1]

        data = self.get_reports_invoices(**kwargs)
        data['apps_dict'] = {app.code: data['apps_dict'][app.code]}
        data['app'] = app
        data.update(self.get_transactions_list(period=period.id))
        data['period'] = period
        alias = app.alias
        data['details_title'] = app.code + (' (%s)' % alias if alias else '')

        return data

    def get_invoices_per_pi(self, **kwargs):
        pi_id = kwargs.get('pi_id', None)

        data = {'pi_id': pi_id,
                'pi_list': [u for u in self.app.dm.get_users() if u.is_pi],
                }

        if pi_id is None:
            return data

        pi_user = self.__get_pi_user(kwargs)
        dm = self.app.dm  # shortcut

        def _filter(b):
            pi = b.owner.get_pi()
            return (b.resource.daily_cost > 0 and
                    b.start <= dm.now() and
                    not b.is_slot and pi and pi.id == pi_user.id)

        entries = []

        for b in dm.get_bookings():
            if _filter(b):
                entries.append({'id': b.id,
                                'title': self.booking_to_event(b)['title'],
                                'date': b.start,
                                'amount': b.total_cost,
                                'type': 'booking'
                                })

        for t in dm.get_transactions():
            if t.user.id == pi_user.id:
                entries.append({'id': t.id,
                                'title': t.comment,
                                'date': t.date,
                                'amount': t.amount,
                                'type': 'transaction'
                                })

        invoice_periods = self.get_invoice_periods_list()['invoice_periods']
        for ip in invoice_periods:
            if ip['order'] > 0:
                entries.append({'id': ip['id'],
                                'title': ip['period'],
                                'date': ip['end'],
                                'amount': 0,
                                'type': 'summary'
                                })

        entries.sort(key=lambda e: e['date'])

        total = 0
        for e in entries:
            if e['type'] == 'summary':
                e['amount'] = total
                if total > 0:
                    total = 0
            else:
                total += e['amount']

        data.update({
            'pi': pi_user,
            'entries': entries,
            'total': total,
            'table_file_prefix': 'all_invoices_PI_' + pi_user.name.replace(' ', '_')
        })

        return data

    def get_invoices_lab_list(self, **kwargs):

        period = self.__get_period(kwargs)
        bookings, range_dict = self.get_booking_in_range(kwargs)

        pi_user = self.__get_pi_user(kwargs)

        apps_dict = {a.id: [] for a in pi_user.get_applications()}
        all_bookings = []
        for b in bookings:
            if b.get('pi_id', None) != pi_user.id:
                continue

            app_id = b.get('app_id', None)

            if app_id not in apps_dict:
                continue

            apps_dict[app_id].append(b)
            all_bookings.append(b)

        result = {
            'pi': pi_user,
            'apps_dict':  apps_dict,
            'all_bookings': all_bookings,
            'total': sum(b['total_cost'] for b in all_bookings)
        }

        result.update(range_dict)
        result.update(self.get_transactions_list(period=period.id,
                                                 pi=pi_user.id))

        return result

    def get_reports_bookings_extracosts(self, **kwargs):
        all_bookings, range_dict = self.get_booking_in_range(kwargs)

        u = self.app.user
        if not u.is_manager:
            raise Exception("Only manager users can access this page")

        bookings = []

        for b in all_bookings:
            if len(b['costs']):
                bookings.append(b)

        q1 = get_quarter()
        q0 = get_quarter(q1[0] - dt.timedelta(days=1))

        self.get_transactions_list(**kwargs)

        result = {
            'bookings': bookings,
            'q0': q0,
            'q1': q1
        }

        result.update(range_dict)

        return result

    def get_booking_costs_table(self, **kwargs):
        resources = self.app.dm.get_resources()

        return {'data': [(r.name, r.status, r.tags) for r in resources]}

    def get_applications_check(self, **kwargs):

        dm = self.app.dm

        sinceArg = kwargs.get('since', None)

        if sinceArg:
            since = datetime_from_isoformat(sinceArg)
        else:
            since = self.app.dm.now() - dt.timedelta(days=183)  # 6 months
        results = {}

        accountsJson = self.app.sll_pm.fetchAccountsJson()
        usersDict = {a['email'].lower(): a for a in accountsJson}

        for application in dm.get_applications():
            app_results = {}
            errors = []

            if application.created < since:
                continue

            orderCode = application.code.upper()
            orderJson = self.app.sll_pm.fetchOrderDetailsJson(orderCode)

            if orderJson is None:
                errors.append('Invalid application ID %s' % orderCode)
            else:
                fields = orderJson['fields']
                pi_list = fields.get('pi_list', [])
                pi_missing = []

                for piTuple in pi_list:
                    piName, piEmail = piTuple
                    piEmail = piEmail.lower()

                    pi = dm.get_user_by(email=piEmail)
                    piInfo = ''
                    if pi is None:
                        if piEmail in usersDict:
                            piInfo = "in the portal, pi: %s" % usersDict[piEmail]['pi']
                        else:
                            piInfo = "NOT in the portal"

                    else:
                        if pi.id != application.creator.id and pi not in application.users:
                            piInfo = 'NOT in APP'
                    if piInfo:
                        pi_missing.append((piName, piEmail, piInfo))

                if pi_missing:
                    app_results['pi_missing'] = pi_missing

            if errors:
                app_results['errors'] = errors

            if app_results:
                app_results['application'] = application
                results[orderCode] = app_results

        return {'checks': results,
                'since': since
                }

    def get_report_pis_usage(self, **kwargs):

        bookings, range_dict = self.get_booking_in_range(kwargs, asJson=False)

        pi_dict = {}
        univ_dict = self.app.dm.get_universities_dict()

        def _get_univ(email, default=None):
            for k, v in univ_dict.items():
                if email.endswith(k):
                    return v
            return default


        for b in bookings:
            pi = b.owner.get_pi()
            if pi:
                parts = pi.name.split()
                first_name = ' '.join(parts[:-1])
                last_name = parts[-1]
                if not pi.email in pi_dict:
                    pi_dict[pi.email] = {
                        'first_name': first_name,
                        'last_name': last_name,
                        'email': pi.email,
                        #'email_rev': pi.email[::-1],  # reverse email for sorting
                        'university': _get_univ(pi.email, 'z-Unknown'),
                        'bookings': 0,
                        'days': 0,
                        'users': set()
                    }
                pi_entry = pi_dict[pi.email]
                pi_entry['bookings'] += 1
                pi_entry['days'] += b.days
                pi_entry['users'].add(b.owner.email)

        data = {
            'pi_list': sorted(pi_dict.values(), key=lambda pi: pi['university'].lower())
        }
        data.update(range_dict)

        return data

    # --------------------- RAW (development) content --------------------------
    def get_raw_booking_list(self, **kwargs):
        bookings = self.app.dm.get_bookings()
        return {'bookings': [self.booking_to_event(b) for b in bookings]}

    def get_raw_applications_list(self, **kwargs):
        return {'applications': self.app.dm.get_visible_applications()}

    def get_forms_list(self, **kwargs):
        return  {'forms': self.app.dm.get_forms()}

    def get_raw_invoice_periods_list(self, **kwargs):
        return {'invoice_periods': self.app.dm.get_invoice_periods()}

    def get_raw_transactions_list(self, **kwargs):
        return {'transactions': self.app.dm.get_transactions()}

    def get_transaction_form(self, **kwargs):
        dm = self.app.dm
        transaction_id = kwargs['transaction_id']
        if transaction_id:
            t = dm.get_transaction_by(id=transaction_id)
        else:
            t = dm.Transaction(date=dt.datetime.now(),
                               amount=0,
                               comment='')

        return {
            'transaction': t,
            'pi_list': [u for u in dm.get_users() if u.is_pi]
        }

    def get_invoice_periods_list(self, **kwargs):
        c = 0
        periods = []

        for ip in self.app.dm.get_invoice_periods(orderBy='start'):
            p = {
                'id': ip.id,
                'status': ip.status,
                'start': ip.start,
                'end': ip.end,
                'period': pretty_quarter((ip.start, ip.end))
            }
            if ip.status != 'disabled':
                c += 1
                p['order'] = c
            else:
                p['order'] = 0
            periods.append(p)

        periods.sort(key=lambda p: p['order'], reverse=True)

        return {'invoice_periods': periods}

    def get_invoice_period_form(self, **kwargs):
        dm = self.app.dm
        invoice_period_id = kwargs['invoice_period_id']
        if invoice_period_id:
            ip = dm.get_invoice_period_by(id=invoice_period_id)
        else:
            ip = dm.InvoicePeriod(status='active',
                                 start=dt.datetime.now(),
                                 end=dt.datetime.now())

        return {
            'invoice_period': ip
        }

    def get_transactions_list(self, **kwargs):
        dm = self.app.dm  # shortcut
        period = dm.get_invoice_period_by(id=int(kwargs['period']))

        def _filter(t):
            return ((period.start < t.date < period.end) and
                    ('pi' not in kwargs or t.user.id == kwargs['pi'] ))

        transactions = [t for t in dm.get_transactions() if _filter(t)]
        transactions_dict = {}

        for t in transactions:
            user_id = t.user.id
            if user_id not in transactions_dict:
                transactions_dict[user_id] = 0
            transactions_dict[user_id] += int(float(t.amount))

        return {
            'transactions': transactions,
            'period': period,
            'transactions_dict': transactions_dict
        }

    def get_invoice_period(self, **kwargs):
        dm = self.app.dm  # shortcut
        period = dm.get_invoice_period_by(id=int(kwargs['period']))
        tabs = [
            {'label': 'overall',
             'template': 'time_distribution.html'
             },
            {'label': 'invoices',
             'template': 'invoices_list.html'
             },
            {'label': 'transactions',
             'template': 'transactions_list.html',
             }
        ]
        tab = kwargs.get('tab', tabs[0]['label'])
        data = {
            'period': period,
            'tabs': tabs,
            'selected_tab': tab,
            'base_url': flask.url_for('main',
                                 content_id='invoice_period',
                                 period=period.id,
                                 tab=tab)
        }

        report_args = {
            'start': pretty_date(period.start),
            'end': pretty_date(period.end),
            'details': kwargs.get('details', None),
            'group': kwargs.get('group', 1)
        }

        data.update(self.get_transactions_list(period=period.id))
        data.update(self.get_reports_invoices(**report_args))
        data.update(self.get_reports_time_distribution(**report_args))

        return data

    def get_projects_list(self, **kwargs):
        # FIXME Define access/permissions for other users
        projects = []
        for p in self.app.dm.get_projects():
            pi = p.user.get_pi()
            if pi:
                apps = pi.get_applications()
                # skip this project from the list if the application is confidential
                # and the user has not access to it
                if apps and not apps[0].allows_access(self.app.user):
                    continue
            projects.append(p)

        return {'projects': projects}

    def get_project_form(self, **kwargs):
        dm = self.app.dm
        project_id = kwargs['project_id']
        if project_id:
            project = dm.get_project_by(id=project_id)
        else:
            user = self.app.user
            now = dm.now()
            project = dm.Project(status='active',
                                 date=now,
                                 last_update_date=now,
                                 last_update_user_id=user.id,
                                 title='',
                                 description='')
            if not self.app.user.is_manager:
                project.creation_user = project.user = user

        return {
            'project': project,
            'possible_owners': self.get_pi_labs()
        }

    def get_project_details(self, **kwargs):
        # FIXME Define access/permissions for other users
        user = self.app.user  # shortchut

        project = self.app.dm.get_project_by(id=kwargs['project_id'])

        if project is None:
            raise Exception("Invalid Project Id %s" % kwargs['project_id'])

        if not user.is_manager and not user.same_pi(project.user):
            raise Exception("You do not have permissions to see this project")


        entries = sorted(project.entries, key=lambda e: e.date, reverse=True)

        return {
            'project': project,
            'entries': entries,
            'entry_types': self.get_entry_types()
        }

    def get_entry_types(self):
        return {
            'grids_preparation':
                {'label': 'Grids Preparation',
                 'group': 1,
                 'iconClass': "fas fa-th fa-inverse",
                 'imageClass': "img--picture",
                 'report': "report_grids_preparation.html"
                 },
            'grids_storage':
                {'label': 'Grids Storage',
                 'group': 1,
                 'iconClass': "fas fa-box fa-inverse",
                 'imageClass': "img--picture",
                 'report': "report_grids_storage.html"
                 },
            'screening':
                {'label': 'Screening',
                 'group': 2,
                 'iconClass': "fas fa-search fa-inverse",
                 'imageClass': "img--location",
                 'report': "report_screening.html"
                 },
            'data_acquisition':
                {'label': 'Data Acquisition',
                 'group': 2,
                 'iconClass': "far fa-image fa-inverse",
                 'imageClass': "img--location",
                 'report': "report_data_acquisition.html"
                 },
            'note':
                {'label': 'Note',
                 'group': 3,
                 'iconClass': "fas fa-sticky-note fa-inverse",
                 'imageClass': "img--picture"
                 },
        }

    def get_entry_form(self, **kwargs):
        dm = self.app.dm
        now = dm.now()
        entry_id = kwargs['entry_id']
        if entry_id:
            entry = dm.get_entry_by(id=entry_id)
            if kwargs.get('copy_entry', False):
                entry.id = None
                entry.title = "Copy of " + entry.title
                entry.creation_date = now
                entry.creation_user_id = self.app.user.id
                entry.last_update_date = now
                entry.last_update_user_id = self.app.user.id
        else:
            project_id = kwargs['entry_project_id']
            project = dm.get_project_by(id=project_id)

            entry = dm.Entry(date=now,
                             creation_date=now,
                             creation_user_id=self.app.user.id,
                             last_update_date=now,
                             last_update_user_id=self.app.user.id,
                             type=kwargs['entry_type'],
                             project=project,
                             title='',
                             description='',
                             extra={})

        entry_type = self.get_entry_types()[entry.type]
        form_id = "entry_form:%s" % entry.type
        form = dm.get_form_by(name=form_id)
        if form:
            self.set_form_values(form, entry.extra.get('data', {}))

        data = {
            'entry': entry,
            'entry_type_label': entry_type['label'],
            'definition': None if form is None else form.definition
        }
        data.update(self.get_grids_storage())

        return data

    def get_entry_report(self, **kwargs):
        dm = self.app.dm
        entry_id = kwargs['entry_id']
        entry = dm.get_entry_by(id=entry_id) if entry_id else None

        if entry is None:
            raise Exception("Please provide a valid Entry id. ")

        entry_type = self.get_entry_types()[entry.type]
        data = entry.extra['data']

        if not 'report' in entry_type:
            raise Exception("There is no Report associated with this Entry. ")

        images = []

        base64 = image.Base64Converter(max_size=(1024, 1024))
        for row in data.get('images_table', []):
            if 'image_file' in row:
                fn = dm.get_entry_path(entry, row['image_file'])
                row['image_data'] = 'data:image/%s;base64, ' + base64.from_path(fn)
                images.append(row)

        # Group data rows by gridboxes (label)
        if entry.type in ['grids_preparation', 'grids_storage']:
            #TODO: Some possible validations
            #TODO:      - There are no more that 4 slots per gridbox
            #TODO:      - There are no duplicated slots
            table = data[entry.type + '_table']
            gridboxes = {}

            for row in table:
                label = row.get('gridbox_label', '')
                if label not in gridboxes:
                    gridboxes[label] = {}
                slots = map(int, row['grid_position'])
                for s in slots:
                    gridboxes[label][s] = row

            data['gridboxes'] = gridboxes

        session = None
        if entry.type == 'data_acquisition':
            session_name = data.get('session_name', '').strip().lower()
            session = dm.get_session_by(name=session_name)

        pi = entry.project.user.get_pi()
        # TODO: We should store some properties in EMhub and avoid this request
        pi_info = self.app.sll_pm.fetchAccountDetailsJson(pi.email) if pi else None

        # Create a default dict based on data to avoid missing key errors in report
        ddata = defaultdict(lambda : 'UNKNOWN')
        ddata.update(data)

        return {
            'entry': entry,
            'entry_type': entry_type,
            'data': ddata,
            'images': images,
            'pi_info': pi_info,
            'session': session
        }

    def get_grids_storage(self, **kwargs):
        return self.get_grids_cane(**kwargs)

    def get_grids_cane(self, **kwargs):
        range = kwargs.get('pucks_range', '1-9999')  # by default all

        min_id, max_id = range.split('-')
        condStr = 'id>=%s and id<=%s' % (min_id, max_id)

        pucks = self.app.dm.get_pucks(condition=condStr, orderBy='id')

        dewars = defaultdict(lambda :defaultdict(dict))

        dewar = int(kwargs.get('dewar', 0) or 0)
        cane = int(kwargs.get('cane', 0) or 0)

        storage = self.app.dm.PuckStorage(pucks)

        for puck in storage.pucks():
            puck['gridboxes'] = defaultdict(dict)

        cond = "type=='grids_storage'"
        for entry in self.app.dm.get_entries(condition=cond):
            table = entry.extra['data']['grids_storage_table']
            for row in table:
                try:
                    slot = int(row['box_position'])
                    puck = storage.get_puck(int(row['puck_id']))
                    slot_key = ','.join(row['grid_position'])
                    row['entry'] = entry
                    puck['gridboxes'][slot][slot_key] = row
                except:
                    pass

        return {
            'storage': storage,
            'pucks_range': range,
            'dewar': storage.get_dewar(dewar),
            'cane': storage.get_cane(dewar, cane)
        }

    def get_grids_puck(self, **kwargs):
        data = self.get_grids_cane(**kwargs)
        data['puck'] = int(kwargs.get('puck', 0) or 0)
        return data

    def get_raw_user_issues(self, **kwargs):
        users = self.get_users_list()['users']
        filterKey = kwargs.get('filter', 'noroles')
        filterName = '_filter_%s' % filterKey

        def _filter_noapp(u):
            """ Users with No Active Application """
            return not u.is_manager and not u.project_codes

        def _filter_noroles(u):
            """ Users with No Roles """
            return not u.is_manager and not u.roles

        _filter = locals()[filterName]

        new_users = [u for u in users if _filter(u)]

        return {'users': new_users,
                'filterDesc': _filter.__doc__
                }

    def get_raw_forms_list(self, **kwargs):
        return {'forms': [
            {'id': f.id,
             'name': f.name,
             'definition': json.dumps(f.definition)
        } for f in self.app.dm.get_forms()]}

    def get_raw_entries_list(self, **kwargs):
        cond = ['%s="%s"' % (k, kwargs[k]) for k in ['id', 'type'] if k in kwargs]
        return {
            'entries': self.app.dm.get_entries(condition=' and '.join(cond))
        }

    def get_raw_pucks_list(self, **kwargs):
        return self.get_grids_cane(**kwargs)

    def get_create_session_form(self, **kwargs):
        dm = self.app.dm  # shortcut
        booking_id = kwargs['booking_id']
        b = dm.get_bookings(condition="id=%s" % booking_id)[0]

        return {
            'booking': b,
            'cameras': dm.get_session_cameras(b.resource.id),
            'processing': dm.get_session_processing(),
            'session_name': dm.get_new_session_info(booking_id)['name']
        }

    # --------------------- Internal  helper methods ---------------------------
    def booking_to_event(self, booking, **kwargs):
        """ Return a dict that can be used as calendar Event object. """
        resource = booking.resource
        # Bookings should have resources, just in case an erroneous one
        if resource is None:
            resource_info = {'id': None, 'name': ''}
        else:
            resource_info = {'id': resource.id, 'name': resource.name,
                             'is_microscope': resource.is_microscope
                             }
        owner = booking.owner
        owner_name = owner.name
        o = booking.operator  #  shortcut
        if o:
            operator_dict = {'id': o.id, 'name': o.name}
        else:
            operator_dict = {'id': None, 'name': ''}

        creator = booking.creator
        a = booking.application
        user = self.app.user
        b_title = booking.title
        b_description = booking.description

        user_can_book = False
        # Define which users are allowed to modify the booking
        # - managers
        # - application creators
        # - the owner and pi of the owner
        can_modify_list = [owner.id]

        if a is not None:
            can_modify_list.append(a.creator.id)

        if user.is_manager and (a is None or a.allows_access(user)):
            can_modify_list.append(user.id)

        pi = owner.get_pi()
        if pi is not None:
            can_modify_list.append(pi.id)

        user_can_modify = user.id in can_modify_list
        user_can_view = user_can_modify or user.same_pi(owner)
        color = resource.color if resource else 'grey'

        application_label = 'None'

        if booking.type == 'downtime':
            color = 'rgba(181,4,0,1.0)'
            title = "%s (DOWNTIME): %s" % (resource.name, b_title)
        if booking.type == 'maintenance' or any(k in b_title for k in ['cycle', 'installation', 'maintenance', 'afis']):
            color = 'rgba(255,107,53,1.0)'
            title = "%s (MAINTENANCE): %s" % (resource.name, b_title)
        elif booking.type == 'slot':
            color = color.replace('1.0', '0.5')  # transparency for slots
            title = "%s (SLOT): %s" % (resource.name,
                                       booking.slot_auth.get('applications', ''))
            user_can_book = user.can_book_slot(booking)
        else:
            # Show all booking information in title in some cases only
            appStr = '' if a is None else ', %s' % a.code
            extra = "%s%s" % (owner.name, appStr)
            if user_can_view:
                title = "%s (%s) %s" % (resource_info['name'], extra, b_title)
                if a:
                    application_label = a.code
                    if a.alias:
                        application_label += "  (%s)" % a.alias
            else:
                title = "%s (%s)" % (resource_info['name'], extra)
                b_title = "Hidden title"
                b_description = "Hidden description"

        bd = {
            'id': booking.id,
            'title': title,
            'description': b_description,
            'start': datetime_to_isoformat(booking.start),
            'end': datetime_to_isoformat(booking.end),
            'color': color,
            'textColor': 'white',
            'resource': resource_info,
            'creator': {'id': creator.id, 'name': creator.name},
            'owner': {'id': owner.id, 'name': owner_name},
            'operator': operator_dict,
            'type': booking.type,
            'booking_title': b_title,
            'user_can_book': user_can_book,
            'user_can_view': user_can_view,
            'user_can_modify': user_can_modify,
            'slot_auth': booking.slot_auth,
            'repeat_id': booking.repeat_id,
            'repeat_value': booking.repeat_value,
            'days': booking.days,
            'experiment': booking.experiment,
            'application_label': application_label,
            'costs': booking.costs,
            'total_cost': booking.total_cost
        }

        if kwargs.get('prettyDate', False):
            bd['pretty_start'] = pretty_datetime(booking.start)
            bd['pretty_end'] = pretty_datetime(booking.end)

        if kwargs.get('piApp', False):
            if pi is not None:
                bd['pi_id'] = pi.id
                bd['pi_name'] = pi.name

            app = booking.application
            if app is not None:
                bd['app_id'] = app.id

        return bd

    def user_profile_image(self, user):
        if getattr(user, 'profile_image', None):
            return flask.url_for('images.user_profile', user_id=user.id)
        else:
            return flask.url_for('images.static', filename='user-icon.png')

    def _get_facility_staff(self, unit):
        """ Return the list of facility personnel.
        First users in the list should  be the facility Head.
        """
        staff = []

        for u in self.app.dm.get_users():
            if u.is_staff and u.staff_unit == unit:
                if 'head' in u.roles:
                    staff.insert(0, u)
                else:
                    staff.append(u)

        return staff

    def _get_display_condition(self):
        """ Compose condition str for the get_sessions query.
        Depending on the user role we show specific sessions only.
        """
        user = self.app.user

        if user.is_manager:
            return None

        condition = 'operator_id == %s' % user.get_id()
        lab_members = user.get_lab_members()
        if user.is_pi and len(lab_members):
            membersId = ",".join(u.get_id() for u in lab_members)
            condition = "operator_id IN (%s)" % membersId

        return condition

    def _get_users_from_portal(self, status=None):
        """ Retrieve users from Portal with a given status.
        If status is None, all will be retrieved.
        """
        dm = self.app.dm
        users = []

        for pu in self.app.sll_pm.fetchAccountsJson():
            user = dm.get_user_by(email=pu['email'])

            if user is None:
                invoiceRef = pu['invoice_ref']

                if pu['status'] == 'enabled':
                    pu['pi_user'] = None

                    if not pu['pi']:
                        pi = dm.get_user_by(email=invoiceRef)
                        if pi is None:
                            pu['status'] = 'error: Missing PI'
                        else:
                            pu['status'] = 'ready: user'
                            pu['pi_user'] = pi
                    else:
                        if invoiceRef.strip():
                            pu['status'] = 'ready: pi'
                        else:
                            pu['status'] = 'error: Missing Invoice Reference'

                    if status is None or pu['status'].startswith(status):
                        users.append(pu)

        return users

    def get_pi_labs(self):
        # Send a list of possible owners of bookings
        # 1) Managers or admins can change the ownership to any user
        # 2) Application managers can change the ownership to any user in their
        #    application
        # 3) Other users can not change the ownership
        user = self.app.user  # shortcut
        if not user.is_authenticated:
            return []

        if user.is_manager:
            piList = [u for u in self.app.dm.get_users() if u.is_pi]
        elif user.is_application_manager:
            apps = [a for a in user.created_applications if a.is_active]
            piSet = {user.get_id()}
            piList = [user]
            for a in apps:
                for pi in a.users:
                    if pi.get_id() not in piSet:
                        piList.append(pi)
        elif user.is_pi:
            piList = [user]
        else:
            piList = []

        def _userjson(u):
            return {'id': u.id, 'name': u.name}

        # Group users by PI
        labs = []
        for u in piList:
            if u.is_pi:
                lab = [_userjson(u)] + [_userjson(u2) for u2 in u.get_lab_members()]
                labs.append(lab)

        if user.is_staff:
            labs.append([_userjson(u) for u in self._get_facility_staff(user.staff_unit)])

        return labs

    def get_possible_operators(self):
        dm = self.app.dm

        if self.app.user.is_authenticated and self.app.user.is_manager:
            return [{'id': u.id, 'name': u.name}
                    for u in dm.get_users() if 'manager' in u.roles]
        return  []

    def get_booking_in_range(self, kwargs, asJson=True):
        """ Return the list of bookings in the given range.
         It will also attach PI information to each booking.
         This function is used from report functions.
         If 'start' and 'end' keys are not in kwargs, the current
         year quarter will be used for the range.
        """

        if 'start' in kwargs and 'end' in kwargs:
            # d = request.json or request.form
            d = {'start': kwargs['start'], 'end': kwargs['end']}
        else:
            # If date range is not passed, let's use by default the
            # current quarter
            now = dt.datetime.now()
            qi = (now.month - 1) // 3
            start, end = [
                ('01/01', '03/31'),
                ('01/04', '06/30'),
                ('01/07', '09/30'),
                ('01/10', '12/31')
            ][qi]
            d = {'start': '%d/%s' % (now.year, start),
                 'end': '%d/%s' % (now.year, end)
                 }

        bookings = self.app.dm.get_bookings_range(
            datetime_from_isoformat(d['start'].replace('/', '-')),
            datetime_from_isoformat(d['end'].replace('/', '-'))
        )

        def process_booking(b):
            if not asJson:
                return b
            return self.app.dc.booking_to_event(b, prettyDate=True, piApp=True)

        def _filter(b):
            return b.resource.daily_cost > 0 and not b.is_slot

        bookings = [process_booking(b) for b in bookings if _filter(b)]

        return bookings, d



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *              Grigory Sharov (gsharov@mrc-lmb.cam.ac.uk) [2]
# *
# * [1] SciLifeLab, Stockholm University
# * [2] MRC Laboratory of Molecular Biology (MRC-LMB)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'delarosatrevin@scilifelab.se'
# *
# **************************************************************************

import os
import shutil
import datetime as dt

from emhub.data import DataManager
from emhub.data.imports import TestDataBase


class TestData(TestDataBase):
    """ Class to create a testing dataset for a given DataManager.
    """
    def _populateUsers(self, dm):
        # Create user table
        usersData = [
            # dev (D)
            ('Don Stairs', 'admin', None),  # 2

            # admin (A)
            ('Anna Mull', 'admin,manager,head', None),   # 3
            ('Arty Ficial', 'admin', None),  # 4

            # managers (M)
            ('Monty Carlo', 'manager', None),  # 5
            ('Moe Fugga', 'manager', None),  # 6

            # pi (P)
            ('Polly Tech', 'pi', None),  # 7
            ('Petey Cruiser', 'pi', None),  # 8
            ('Pat Agonia', 'pi', None),  # 9
            ('Paul Molive', 'pi', None),  # 10
            ('Pat Ernity', 'pi', None),  # 11

            # users (R, S)
            ('Ray Cyst', 'user', 7),  # 12
            ('Rick Shaw', 'user', 7),  # 13
            ('Rachel Slurs', 'user', 7),  # 14

            ('Reggie Stration', 'user', 8),  # 15
            ('Reuben Sandwich', 'user', 8),  # 16
            ('Sara Bellum', 'user', 8),  # 17
            ('Sam Owen', 'user', 8),  # 18

            ('Sam Buca', 'user', 10),  # 19
            ('Sarah Yevo', 'user', 10),  # 20
            ('Sven Gineer', 'user', 10),  # 21
            ('Sharon Needles', 'user', 10),  # 22

            ('Ray Diation', 'user', 11),  # 22
            ('Sal Ami', 'user', 11)   # 23
        ]

        for name, roles, pi in usersData:
            first, last = name.lower().split()
            roles = roles.split(',')
            dm.create_user(username=last,
                           email='%s.%s@emhub.org' % (first, last),
                           phone='%d-%d%d' % (len(roles), len(first), len(last)),
                           password=last,
                           name=name,
                           roles=roles,
                           pi_id=pi)

    def _populateApplications(self, dm):
        templateInfo = [
            {'title': 'BAG Application Form - 2019/2020',
             'description': 'Information required in order to submit a request '
                            'for group allocation time for one year.',
             'status': 'active',
             },
            {'title': 'Rapid Access for SPA and Tomography',
             'description': 'Application Form required to apply for time on a '
                            'project for either screening or data-collection. '
                            'Machine time will be allocate on 24 or 48 hours '
                            'slots. ',
             'status': 'active',
             },
            {'title': 'BAG Application Form - 2020/2021',
             'description': 'Information required in order to submit a request '
                            'for group allocation time for one year.',
             'status': 'closed'
             },
            {'title': 'Single Particle Application 161026',
             'description': 'Information required to request time for one year.',
             'status': 'rejected',
             },
            {'title': 'Internal Users Template',
             'description': 'This is a special template used for internal users. ',
             'status': 'rejected',
             },
        ]

        templates = [dm.create_template(**ti) for ti in templateInfo]

        applications = [
            {'code': 'CEM00297',
             'alias': 'BAG Lund',
             'status': 'active',
             'title': 'Bag Application for Lund University 2019/20',
             'creator': 'agonia',
             'template_id': templates[0].id,
             'invoice_reference': 'AAA',
             'invoice_address': '',
             'resource_allocation': {'quota': {'talos': 10, 'krios': 5},
                                     'noslot': []},
             'description': "Current application BAG for Lund University."
             },
            {'code': 'CEM00315',
             'alias': 'BAG SU',
             'status': 'active',
             'title': 'Bag Application for Stockholm University',
             'description': '',
             'creator': 'tech',
             'template_id': templates[0].id,
             'invoice_reference': 'BBB',
             'invoice_address': '',
             'resource_allocation': {'quota': {'talos': 10, 'krios': 5},
                                     'noslot': []}
             },
            {'code': 'CEM00332',
             'alias': 'RAA Andersson',
             'status': 'active',
             'title': 'Rapid Access application',
             'description': '',
             'creator': 'ernity',
             'template_id': templates[1].id,
             'invoice_reference': 'ZZZ',
             'invoice_address': '',
             'resource_allocation': {'quota': {'talos': 2, 'krios': 1},
                                     'noslot': []}
             },
            {'code': 'DBB00001',
             'alias': 'SU-DBB',
             'status': 'active',
             'title': 'Internal DBB project',
             'description': '',
             'creator': 'tech',
             'template_id': templates[-1].id,
             'invoice_reference': 'DDD',
             'invoice_address': '',
             'resource_allocation': {'quota': {},
                                     'noslot': [1, 2]}
             },
            {'code': 'CEM00345',
             'alias': 'BAG Lund 2021',
             'status': 'review',
             'title': 'Bag Application for Lund University 2021',
             'description': '',
             'creator': 'agonia',
             'template_id': templates[0].id,
             'invoice_reference': 'BBB',
             'invoice_address': '',
             'resource_allocation': {'quota': {'talos': 10, 'krios': 5},
                                     'noslot': []}
             },
            {'code': 'CEM00346',
             'alias': 'BAG SU 2021',
             'status': 'review',
             'title': 'Bag Application for Stockholm University 2021',
             'description': '',
             'creator': 'tech',
             'template_id': templates[0].id,
             'invoice_reference': 'BBB',
             'invoice_address': '',
             'resource_allocation': {'quota': {'talos': 10, 'krios': 5},
                                     'noslot': []}
             },
        ]

        for pDict in applications:
            username = pDict.pop('creator')
            u = dm.get_user_by(username=username)
            pDict['creator_id'] = u.id
            dm.create_application(**pDict)

        def __addPi(appCode, piUser):
            a1 = dm.get_application_by(code=appCode)
            u1 = dm.get_user_by(username=piUser)
            a1.users.append(u1)

        for u in ['cruiser', 'agonia']:
            __addPi('DBB00001', u)

        for u in ['cruiser']:
            __addPi('CEM00315', u)

        for u in ['molive']:
            __addPi('CEM00297', 'ernity')

        dm.commit()

    def __firstMonday(self, now):
        td = dt.timedelta(days=now.weekday())
        td7 = dt.timedelta(days=7)

        prevMonday = now - td

        while prevMonday.month == now.month:
            prevMonday = prevMonday - td7

        return prevMonday + td7


    def _populateBookings(self, dm):
        now = dm.now().replace(minute=0, second=0)
        month = now.month
        self.firstMonday = self.__firstMonday(now)
        td = dt.timedelta  # shortcut

        def fm(shift):
            return (self.firstMonday + td(days=shift)).replace(hour=9)

        # Create a booking at the downtime from today to one week later
        for r in [1, 3]:  # Krios 1 and  Talos
            for s in [0, 14]:
                dm.create_booking(title='',
                                  start=fm(s),
                                  end=fm(s+5).replace(hour=23),
                                  type='slot',
                                  slot_auth={'applications': ['CEM00297', 'CEM00315']},
                                  resource_id=r,
                                  creator_id=2,  # first user for now
                                  owner_id=2,  # first user for now
                                  description="Slot for National BAGs")

        # Create a downtime from today to one week later
        dm.create_booking(title='',
                          start=fm(17),
                          end=fm(21).replace(hour=23),
                          type='downtime',
                          resource_id=1,
                          creator_id=2,  # first user for now
                          owner_id=2,  # first user for now
                          description="Some downtime for some problem")

        dm.create_booking(title='',
                          start=fm(-7),
                          end=fm(-6).replace(hour=23),
                          type='booking',
                          resource_id=1,
                          creator_id=3,  # ann mull
                          owner_id=3,  # mull
                          description="")

        # Create booking for normal user
        dm.create_booking(title='',
                          start=fm(0),
                          end=fm(1).replace(hour=23),
                          type='booking',
                          resource_id=1,
                          creator_id=13,  # first user for now
                          owner_id=13,  # first user for now
                          description="")

        dm.create_booking(title='',
                          start=fm(2),
                          end=fm(4).replace(hour=23),
                          type='booking',
                          resource_id=1,
                          creator_id=2,  # first user for now
                          owner_id=16,  # Sara Belum
                          description="Krios 2 for user 3")

        dm.create_booking(title='Slot 2: RAPID',
                      start=fm(21),
                      end=fm(22).replace(hour=23),
                      type='slot',
                      slot_auth={'applications': ['CEM00332']},
                      resource_id=3,
                      creator_id=2,  # first user for now
                      owner_id=2,  # first user for now
                      description="Talos slot for RAPID applications")

        # create a repeating event
        dm.create_booking(title='Dropin',
                          start=fm(2),
                          end=fm(2).replace(hour=16),
                          type='slot',
                          repeat_value='bi-weekly',
                          repeat_stop=now + dt.timedelta(2*30),
                          resource_id=7,
                          creator_id=2,  # first user for now
                          owner_id=2,  # first user for now
                          description="Recurrent bi-weekly DROPIN slot. ")

    def _populateSessions(self, dm):
        td = os.environ.get('EMHUB_TESTDATA')
        inst = os.environ.get('EMHUB_INSTANCE')

        dm.create_session(
            name='supervisor_23423452_20201223_123445',
            start=self._datetime(2020, 3, 5, 12, 30, 10),
            end=None,
            status='running',
            #data_path=os.path.join(td, 'hdf5/20181108_relion30_tutorial.h5'),
            acquisition={'voltage': 300,
                         'cs': 2.7,
                         'phasePlate': False,
                         'detector': 'Falcon2',
                         'detectorMode': 'Linear',
                         'pixelSize': 1.1,
                         'dosePerFrame': 1.0,
                         'totalDose': 35,
                         'exposureTime': 1.2,
                         'numOfFrames': 48,
                         },
            stats={'numMovies': 423,
                   'numMics': 0,
                   'numCtf': 0,
                   'numPtcls': 0,
                   },
            resource_id=1,  # Krios 1
            booking_id=None,
            operator_id=1,  # User  X
        )

        shutil.copyfile(os.path.join(td, 'hdf5/20181108_relion30_tutorial.h5'),
                        os.path.join(inst, 'sessions/session_000001.h5'))

        dm.create_session(
            name='epu-mysession_20122310_234542',
            start=self._datetime(2020, 4, 5, 12, 30, 10),
            end=None,
            status='failed',
            #data_path=os.path.join(td, 'hdf5/t20s_pngs.h5'),
            acquisition={'voltage': 300,
                         'cs': 2.7,
                         'phasePlate': False,
                         'detector': 'Falcon2',
                         'detectorMode': 'Linear',
                         'pixelSize': 1.1,
                         'dosePerFrame': 1.0,
                         'totalDose': 35,
                         'exposureTime': 1.2,
                         'numOfFrames': 48,
                         },
            stats={'numMovies': 234,
                   'numMics': 234,
                   'numCtf': 234,
                   'numPtcls': 2,
                   },
            resource_id=2,  # Krios 2
            booking_id=None,
            operator_id=6,  # User  6
        )

        shutil.copyfile(os.path.join(td, 'hdf5/t20s_pngs.h5'),
                        os.path.join(inst, 'sessions/session_000002.h5'))

        dm.create_session(
            name='session_very_long_name',
            start=self._datetime(2020, 5, 7, 12, 30, 10),
            end=self._datetime(2020, 5, 8, 9, 30, 10),
            status='finished',
            data_path=os.path.join(td, 'non-existing-file'),
            acquisition={'voltage': 300,
                         'cs': 2.7,
                         'phasePlate': False,
                         'detector': 'Falcon2',
                         'detectorMode': 'Linear',
                         'pixelSize': 1.1,
                         'dosePerFrame': 1.0,
                         'totalDose': 35,
                         'exposureTime': 1.2,
                         'numOfFrames': 48,
                         },
            stats={'numMovies': 2543,
                   'numMics': 2543,
                   'numCtf': 2543,
                   'numPtcls': 2352534,
                   },
            resource_id=3,  # Talos
            booking_id=None,
            operator_id=12,  # User  12
        )


if __name__ == '__main__':
    instance_path = os.path.abspath(os.environ.get("EMHUB_INSTANCE",
                                                   'instance'))

    if not os.path.exists(instance_path):
        raise Exception("Instance folder '%s' not found!!!" % instance_path)

    dm = DataManager(instance_path, cleanDb=True)
    TestData(dm)